import os


def get_file_path(directories, name):
    """
    Finds the first matching file path within directories.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    for path in directories:
        path = os.path.join(path, name)
        if os.path.isfile(path):
            return path


def get_folder_path(directories, name):
    """
    Finds the first matching folder path within directories.

    :param name: Name of the folder
    :type name: str
    :return: Found folder path
    :rtype: str or None
    """
    for path in directories:
        path = os.path.join(path, name)
        if os.path.isdir(path):
            return path


def get_tool_path(directories, name):
    """
    Finds the first matching executable file within directories.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    for path in directories:
        path = os.path.join(path, name)
        if os.path.isfile(path) and os.access(path, os.X_OK):
            return path


# NOTE(ivasilev) You can use the functions below standalone with
# repository = load_repositories_from('repo_path', '/etc/leapp/repo.d/', manager=None)
def get_common_file_path(repository, name):
    return get_file_path(repository.files, name)


def get_common_folder_path(repository, name):
    return get_folder_path(repository.files, name)


def get_common_tool_path(repository, name):
    return get_tool_path(repository.files, name)



import base64
import copy
import datetime
import json

import six


class ModelViolationError(Exception):
    """
    ModelViolationError is raised if the data in the instances is not matching its definition.
    """


class ModelMisuseError(Exception):
    """
    ModelMisuseError is raised if the Model definition is not valid.
    """


class Field(object):
    """
    Field is the base of all supported fields.
    """
    @property
    def help(self):
        """
        :return: Documentation help string defining what the field is about.
        """
        return self._help or 'No documentation provided for this field `{}`'.format(type(self).__name__)

    def as_nullable(self):
        """
        Set object` "_nullable_" field to True and return the object back
        """
        self._nullable = True
        return self

    def __init__(self, default=None, help=None):  # noqa; pylint: disable=redefined-builtin
        """
        :param default: Default value to be used if the field is not set
        :param help: Documentation string for generating model documentation
        :type help: str
        """
        self._help = help
        self._nullable = False
        self._default = default

        if type(self) == Field:
            raise ModelMisuseError("Do not use this type directly.")

    def _validate_model_value(self, value, name):
        """
        Validates the value in the Model representation

        :param value: Value to be checked
        :param name: Name of the field (used for a better error reporting only)
        :return: None
        """
        if value is None and not self._nullable:
            raise ModelViolationError('The value of "{name}" field is None, but this is not allowed'.format(name=name))

    def _validate_builtin_value(self, value, name):
        """
        Validates the value in the builtin representation

        :param value: Value to be checked
        :param name: Name of the field (used for a better error reporting only)
        :return: None
        """
        if value is None and not self._nullable:
            raise ModelViolationError('The value of "{name}" field is None, but this is not allowed'.format(name=name))

    def _convert_to_model(self, value, name):
        """
        Performs the conversion from a builtin type to the model representation

        :param value: Value to convert
        :param name: Name of the field (used for a better error reporting only)
        :return: Converted value in the model format
        """
        self._validate_builtin_value(value=value, name=name)
        return value

    def _convert_from_model(self, value, name):
        """
        Performs the conversion from a model type to the builtin representation

        :param value: Value to convert
        :param name: Name of the field (used for a better error reporting only)
        :return: Converted value in the builtin format
        """
        self._validate_model_value(value=value, name=name)
        return value

    def from_initialization(self, source, name, target):
        """
        Assigns the value to the target model passed through during the model initialization

        :param source: Dictionary to extract the value from (usually kwargs)
        :type source: dict
        :param name: Name of the field (used for a better error reporting only)
        :type name: str
        :param target: Target model instance
        :type target: Instance of a Model derived class
        :return: None
        """
        # copy the default value for it not to be shared with other
        # instances of the same model
        source_value = copy.copy(source.get(name, self._default))
        self._validate_model_value(value=source_value, name=name)
        setattr(target, name, source_value)

    def to_model(self, source, name, target):
        """
        Converts the value with the given name to the model representation and assigns the attribute

        :param source: Dictionary to extract the value from
        :type source: dict
        :param name: Name of the field (used for a better error reporting only)
        :type name: str
        :param target: Target model instance
        :type target: Instance of a Model derived class
        :return: None
        """
        # copy the default value for it not to be shared with other
        # instances of the same model
        source_value = copy.copy(source.get(name, self._default))
        if source_value is not None:
            source_value = self._convert_to_model(source_value, name=name)
        self._validate_model_value(value=source_value, name=name)
        setattr(target, name, source_value)

    def to_builtin(self, source, name, target):
        """
        Converts the value with the given name to the builtin representation and assigns the field

        :param source: Source model to get the value from
        :type source: Instance of a Model derived class
        :param name: Name of the field (used for a better error reporting only)
        :type name: str
        :param target: Dictionary to set the value to
        :type target: dict
        :return: None
        """
        target[name] = self._convert_from_model(getattr(source, name, None), name=name)

    def serialize(self):
        """
        :return: Serialized form of the workflow
        """
        return {
            'nullable': self._nullable,
            'class_name': type(self).__name__,
            'default': self._default,
            'help': self._help
        }


class BuiltinField(Field):
    """
    Base class for all builtin types to act as pass-through with an additional validation
    """

    @property
    def _model_type(self):
        """
        :return: Returns the type to be used as a model type representation
        """
        raise NotImplementedError("_model_type needs to be overridden")

    @property
    def _builtin_type(self):
        """
        :return: Returns the type to be used as a builtin type representation (e.g. string)
        """
        return self._model_type

    def _validate_model_value(self, value, name):
        super(BuiltinField, self)._validate_model_value(value, name)
        self._validate(value=value, name=name, expected_type=self._model_type)

    def _validate_builtin_value(self, value, name):
        super(BuiltinField, self)._validate_builtin_value(value, name)
        self._validate(value=value, name=name, expected_type=self._builtin_type)

    def _validate(self, value, name, expected_type):
        if not isinstance(expected_type, tuple):
            expected_type = (expected_type,)
        if value is None and self._nullable:
            return
        if not any(isinstance(value, t) for t in expected_type):
            names = ', '.join(['{}'.format(t.__name__) for t in expected_type])
            raise ModelViolationError("Fields {} is of type: {} expected: {}".format(name, type(value).__name__,
                                                                                     names))


class Boolean(BuiltinField):
    """
    Boolean field
    """
    @property
    def _model_type(self):
        return bool


class Float(BuiltinField):
    """
    Float field
    """
    @property
    def _model_type(self):
        return float


class Integer(BuiltinField):
    """
    Integer field (int, long in python 2, int in python 3)
    """
    @property
    def _model_type(self):
        return six.integer_types


class Number(BuiltinField):
    """
    Combined Integer and Float field
    """
    @property
    def _model_type(self):
        return six.integer_types + (float,)


class String(BuiltinField):
    """
    String field
    """
    @property
    def _model_type(self):
        return six.string_types + (six.binary_type,)


class Blob(BuiltinField):
    """
    Blob field
    """
    @property
    def _model_type(self):
        # NOTE(ivasilev) Both string_types and binary_types are necessary here to pass value checks on
        # serialization/deserialization. Serialized JSON-friendly model will have a string type field (the base64
        # ascii string) while deserialized model will have a binary type field.
        return six.string_types + (six.binary_type,)

    def _convert_to_model(self, value, name):
        self._validate_model_value(value=value, name=name)
        if value is None:
            return None

        return base64.b64decode(value)

    def _convert_from_model(self, value, name):
        self._validate_model_value(value=value, name=name)
        if value is None:
            return None

        # NOTE(ivasilev) b64 encoding is always ascii, thus ascii decoding to get a string
        return base64.b64encode(value).decode('ascii')


class DateTime(BuiltinField):
    """
    DateTime field to handle datetime objects which are converted to the ISO format and parsed back from there
    """
    @property
    def _model_type(self):
        return datetime.datetime

    @property
    def _builtin_type(self):
        return six.string_types

    def _convert_to_model(self, value, name):
        self._validate_builtin_value(value=value, name=name)

        if value is None:
            return value

        # We want Z to be appended but it needs support from our side here:
        value = value.rstrip('Z')

        # If there are fractions, use them
        fractions = ''
        if '.' in value:
            fractions = '.%f'

        # Try to parse with timezone specification, else retry without
        try:
            return datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S{fractions}%Z'.format(fractions=fractions))
        except ValueError:
            try:
                return datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S{fractions}'.format(fractions=fractions))
            except ValueError:
                raise ModelViolationError(
                    "The {name} field contains an invalid datetime value: '{value}'".format(name=name, value=value))

    def _convert_from_model(self, value, name):
        self._validate_model_value(value=value, name=name)

        if not value:
            return value

        if not value.utcoffset():
            return value.isoformat() + 'Z'


class EnumMixin(Field):
    """
    EnumMixin adds the ability to use the field as an Enum type of the field
    """
    def __init__(self, choices, **kwargs):
        """
        :param choices: List of values that are allowed for this field
        :type choices: List or tuple of allowed values
        """
        super(EnumMixin, self).__init__(**kwargs)
        if not isinstance(choices, (tuple, list)):
            raise ModelMisuseError("Choices needs to be a list or a tuple")
        self._choices = choices

    def _validate_model_value(self, value, name):
        super(EnumMixin, self)._validate_model_value(value, name)
        self._validate_choices(value, name)

    def _validate_builtin_value(self, value, name):
        super(EnumMixin, self)._validate_builtin_value(value, name)
        self._validate_choices(value, name)

    def _validate_choices(self, value, name):
        if value is not None and value not in self._choices:
            values = ", ".join(map(str, self._choices))
            raise ModelViolationError(
                'The value of "{name}" field must be one of "{values}"'.format(name=name, values=values))

    def serialize(self):
        result = super(EnumMixin, self).serialize()
        result.update({
            'choices': self._choices
        })
        return result


def Nullable(elem_field):
    """
    Helper function to make a field nullable
    """
    return elem_field.as_nullable()


class StringEnum(EnumMixin, String):
    """
    Field that represents an enumeration of Strings
    """


class IntegerEnum(EnumMixin, Integer):
    """
    Field that represents an enumeration of Integers
    """


class FloatEnum(EnumMixin, Float):
    """
    Field that represents an enumeration of Floats
    """


class NumberEnum(EnumMixin, Number):
    """
    Field that represents an enumeration of Numbers
    """


class List(Field):
    """
        List represents lists of `elem_field` values
    """
    def __init__(self, elem_field, minimum=None, maximum=None, **kwargs):
        """
        :param elem_field:
        :type elem: Instance of :py:class:`Field`
        :param minimum: Minimal number of elements
        :type minimum: int or None
        :param maximum: Maximum number of elements
        :type maximum: int or None
        :param default: Default value to use if the field is not set
        :type default: A list of elements with the value type as specified in `elem_field`
        :param help: Documentation string for generating model documentation
        :type help: str
        """
        super(List, self).__init__(**kwargs)
        # We do a copy of the data in default, to avoid some unwanted side effects
        # Comparison to None is necessary
        if self._default is not None:
            self._default = copy.copy(self._default)
        if not isinstance(elem_field, Field):
            raise ModelMisuseError("elem_field must be an instance of a type derived from Field")
        self._elem_type = elem_field
        self._minimum = minimum or 0
        self._maximum = maximum

    def _validate_count(self, value, name):
        message = 'Element count error for field {name} expected between {minimum} and {maximum} elements got {count}'
        count = len(value)
        if not self._minimum <= count <= (self._maximum or count):
            raise ModelViolationError(
                message.format(name=name, minimum=self._minimum, maximum=self._maximum or count, count=count))

    def _validate_model_value(self, value, name):
        super(List, self)._validate_model_value(value, name)
        if isinstance(value, (list, tuple)):
            self._validate_count(value, name)
            for idx, entry in enumerate(value):
                self._elem_type._validate_model_value(entry, name='{}[{}]'.format(name, idx))
        elif value is not None:
            raise ModelViolationError('Expected list but got {} for the {} field'.format(type(value).__name__, name))

    def _validate_builtin_value(self, value, name):
        super(List, self)._validate_builtin_value(value, name)
        if isinstance(value, (list, tuple)):
            self._validate_count(value, name)
            for idx, entry in enumerate(value):
                self._elem_type._validate_builtin_value(entry, name='{}[{}]'.format(name, idx))
        elif value is not None:
            raise ModelViolationError('Expected list but got {} for the {} field'.format(type(value).__name__, name))

    def _convert_to_model(self, value, name):
        self._validate_builtin_value(value=value, name=name)
        if value is None:
            return value
        converter = self._elem_type._convert_to_model
        return list(converter(entry, name='{}[{}]'.format(name, idx)) for idx, entry in enumerate(value))

    def _convert_from_model(self, value, name):
        self._validate_model_value(value=value, name=name)
        if value is None:
            return value
        converter = self._elem_type._convert_from_model
        return list(converter(entry, name='{}[{}]'.format(name, idx)) for idx, entry in enumerate(value))

    def serialize(self):
        result = super(List, self).serialize()
        result.update({
            'element': self._elem_type.serialize(),
            'maximum': self._maximum,
            'minimum': self._minimum
        })
        return result


class Model(Field):
    """
    Model is used to use other Models as fields
    """
    def __init__(self, model_type, **kwargs):
        """
        :param model_type: A :py:class:`leapp.model.Model` derived class
        :type model_type: :py:class:`leapp.model.Model` derived class
        :param help: Documentation string for generating the model documentation
        :type help: str
        """
        super(Model, self).__init__(**kwargs)
        from leapp.models import Model as ModelType  # pylint: disable=import-outside-toplevel
        if not isinstance(model_type, type) or not issubclass(model_type, ModelType):
            raise ModelMisuseError("{} must be a type derived from Model".format(model_type))
        self._model_type = model_type

    def _validate_model_value(self, value, name):
        super(Model, self)._validate_model_value(value, name)
        resolved_model = getattr(self._model_type, '_resolved', self._model_type)
        if value and not isinstance(value, resolved_model):
            raise ModelViolationError('Expected an instance of {} for the {} attribute but got {}'.format(
                resolved_model.__name__, name, type(value)))

    def _validate_builtin_value(self, value, name):
        super(Model, self)._validate_model_value(value, name)
        if value and not isinstance(value, dict):
            raise ModelViolationError('Expected a value for the {} field and got {}'.format(name, type(value).__name__))

    def _convert_to_model(self, value, name):
        self._validate_builtin_value(value, name)
        if value is None:
            return value
        return self._model_type(init_method='to_model', **value)

    def _convert_from_model(self, value, name):
        self._validate_model_value(value, name)
        if value is None:
            return value
        return value.dump()

    def serialize(self):
        result = super(Model, self).serialize()
        result.update({
            'model': self._model_type.__name__
        })
        return result


class JSON(String):
    """
    The JSON field allows to use json encodable python types as a value.

    The value will be internally encoded to a JSON string and converted back into, whatever the result of json.loads
    is for that value passed.

        Note: The value `None`, however follows the same rules as for all fields and requires the field to be nullable,
              to allow this value. Within nested values such as lists or dicts, a None value is perfectly valid.
    """

    @property
    def _model_type(self):
        return six.integer_types + (float, tuple, dict, list) + six.string_types

    def _convert_from_model(self, value, name):
        if value is None:
            if not self._nullable:
                raise ModelViolationError(
                    'The value of "{name}" field is None, but this is not allowed'.format(name=name))
            return value
        try:
            return json.dumps(value, sort_keys=True)
        except (TypeError, ValueError):
            raise ModelViolationError('Expected a json encodable value for the field {}'.format(name))

    def _convert_to_model(self, value, name):
        self._validate_builtin_value(value, name)
        if value is None:
            return value
        return json.loads(value)



import re
import socket


def decorate_getfqdn(fn):
    def wrap(*args, **kwargs):
        result = fn(*args, **kwargs)
        # check whether FQDN conforms to standard, see HOSTNAME(7)
        pattern = r"^(([a-z0-9]|[a-z0-9][a-z0-9\-]*[a-z0-9])\.)*([a-z0-9]|[a-z0-9][a-z0-9\-]*[a-z0-9])$"
        if re.match(pattern, result, re.IGNORECASE):
            return result
        return 'invalid.hostname'

    return wrap


def apply_workaround():
    setattr(socket, 'getfqdn', decorate_getfqdn(socket.getfqdn))



from logging import getLogger
import os
import pkgutil
import sys

from leapp.exceptions import RepoItemPathDoesNotExistError, UnsupportedDefinitionKindError
from leapp.models import get_models, resolve_model_references
import leapp.libraries.common  # noqa # pylint: disable=unused-import
from leapp.repository.actor_definition import ActorDefinition
from leapp.repository.definition import DefinitionKind
from leapp.tags import get_tags
from leapp.topics import get_topics
from leapp.utils.libraryfinder import LeappLibrariesFinder
from leapp.utils.repository import get_repository_id, get_repository_links, get_repository_name
import leapp.workflows
from leapp.workflows import get_workflows


class _LoadStage(object):
    INITIAL = 'initial'
    MODELS = 'models'
    LIBRARIES = 'libraries'
    ACTORS = 'actors'
    WORKFLOWS = 'workflows'


class Repository(object):
    """
    The Repository class represents a place where all resources (actors, models, tags, etc.) are defined. See the
    :doc:`Repository Directory Layout <../repository-dir-layout>`.
    """

    def __init__(self, directory):
        """
        :param directory: Path to the repository folder
        :type directory: str
        """
        self.name = get_repository_name(directory)
        self.log = getLogger('leapp.repository').getChild(self.name)
        self._repo_dir = directory
        self._repo_id = get_repository_id(directory)
        self._repo_links = get_repository_links(directory)
        self._definitions = {}
        self.log.info("A new repository '%s' is initialized at %s", self.name, directory)

    @property
    def repo_dir(self):
        return self._repo_dir

    @property
    def repo_id(self):
        return self._repo_id

    @property
    def repo_links(self):
        return self._repo_links

    def lookup_actor(self, name):
        """
        Finds an actor in the repository

        :param name: Name of the actor
        :type name: str
        :return: None or Actor
        """
        # TODO: what's the behaviour in case of multiple actors of the same name?
        name = name.lower()
        for actor in self.actors:
            actor_name = actor.name.lower()
            actor_class = actor.class_name.lower()
            if name in (actor_name, actor_class):
                return actor
        return None

    @staticmethod
    def lookup_workflow(name):
        """
        Finds a workflow in the repository

        :param name: Name of the workflow class name, Workflow.name, or Workflow.short_name
        :type name: str
        :return: None or Workflow
        """
        name = name.lower()
        for workflow in leapp.workflows.get_workflows():
            workflow_name = workflow.name.lower()
            workflow_class = workflow.__name__.lower()
            workflow_short_name = workflow.short_name
            if name in (workflow_name, workflow_class, workflow_short_name):
                return workflow
        return None

    def add(self, kind, item):
        """
        Adds any supported kind of a resource to the repository

        :param kind: specific kind of the repository resource
        :type kind: :py:class:`leapp.repository.definition.DefinitionKind`
        :param item: Item that will be added
        :type item: :py:class:`leapp.repository.actor_definition.ActorDefiniton` or str
        """
        self.log.debug('Adding %s - %s', kind.name, item.directory if isinstance(item, ActorDefinition) else item)
        if kind not in DefinitionKind.REPO_WHITELIST:
            raise UnsupportedDefinitionKindError('Repositories do not support {kind}.'.format(kind=kind.name))

        # Except for ActorDefinitions all items added are paths and are supposed to be relative paths to the repository
        if not isinstance(item, ActorDefinition):
            full_path = os.path.join(self._repo_dir, item)
            if not os.path.exists(full_path):
                self.log.error("Attempted to add %s, which is not in the repositories path", item)
                raise RepoItemPathDoesNotExistError(kind, item, full_path)
            item = full_path

        self._definitions.setdefault(kind, []).append(item)

    def load(self, resolve=True, stage=None, skip_actors_discovery=False):
        """
        Loads the repository resources

        :param resolve: Decides whether or not to perform the resolving of model references
        :type resolve: bool
        :param stage: Stage to load - Required for repository managers
        :type stage: _LoadStage value
        """
        if not stage or stage is _LoadStage.INITIAL:
            self.log.debug("Loading repository %s", self.name)
            self.log.debug("Loading tag modules")
            self._load_modules(self.tags, 'leapp.tags')
            self.log.debug("Loading topic modules")
            self._load_modules(self.topics, 'leapp.topics')

        if not stage or stage is _LoadStage.MODELS:
            self.log.debug("Loading model modules")
            self._load_modules(self.models, 'leapp.models')
            if resolve:
                resolve_model_references()

        if not stage or stage is _LoadStage.LIBRARIES:
            self.log.debug("Extending PATH for common tool paths")
            self._extend_environ_paths('PATH', self.tools)
            self.log.debug("Extending LEAPP_COMMON_TOOLS for common tool paths")
            self._extend_environ_paths('LEAPP_COMMON_TOOLS', self.tools)
            self.log.debug("Extending LEAPP_COMMON_FILES for common file paths")
            self._extend_environ_paths('LEAPP_COMMON_FILES', self.files)
            self.log.debug("Installing repository provided common libraries loader hook")
            sys.meta_path.append(LeappLibrariesFinder(module_prefix='leapp.libraries.common', paths=self.libraries))
            sys.meta_path.append(LeappLibrariesFinder(module_prefix='leapp.workflows.api', paths=self.apis))

        if not skip_actors_discovery:
            if not stage or stage is _LoadStage.ACTORS:
                self.log.debug("Running actor discovery")
                for actor in self.actors:
                    actor.discover()

        if not stage or stage is _LoadStage.WORKFLOWS:
            self.log.debug("Loading workflow modules")
            self._load_modules(self.workflows, 'leapp.workflows')

    def _load_modules(self, modules, prefix):
        """
        :param modules: Paths to modules of the same type to be imported
        :type list of str
        :param prefix: A prefix to be appended to module name in sys.modules
        :type str
        """
        directories = [os.path.join(self._repo_dir, os.path.dirname(module)) for module in modules]
        prefix = prefix + '.' if not prefix.endswith('.') else prefix
        for importer, name, ispkg in pkgutil.iter_modules(directories, prefix=prefix):
            importer.find_module(name).load_module(name)

    def serialize(self):
        """
        :return: Dictionary of all repository resources
        """
        def filtered_serialization(func, modules):
            result = []
            data = func()
            for entry in data:
                if os.path.abspath(sys.modules[entry.__module__].__file__.replace('.pyc', '.py')) in modules:
                    result.append(entry.serialize())
            return result

        def mapped_actor_data(data):
            data.update({
                'consumes': [model.__name__ for model in data.get('consumes', ())],
                'produces': [model.__name__ for model in data.get('produces', ())],
                'tags': [tag.__name__ for tag in data.get('tags', ())]
            })
            return data

        return {
            'repo_dir': self._repo_dir,
            'actors': [mapped_actor_data(a.serialize()) for a in self.actors],
            'apis': [dict([('path', path)]) for path in self.relative_paths(self.apis)],
            'topics': filtered_serialization(get_topics, self.topics),
            'models': filtered_serialization(get_models, self.models),
            'tags': filtered_serialization(get_tags, self.tags),
            'workflows': filtered_serialization(get_workflows, self.workflows),
            'tools': [dict([('path', path)]) for path in self.relative_paths(self.tools)],
            'files': [dict([('path', path)]) for path in self.relative_paths(self.files)],
            'libraries': [dict([('path', path)]) for path in self.relative_paths(self.libraries)]
        }

    def _extend_environ_paths(self, name, paths):
        os.environ[name] = ':'.join(os.environ.get(name, '').split(':') + list(
            os.path.join(self._repo_dir, path) for path in paths))

    def relative_paths(self, paths):
        """
        :return: Tuple of repository relative paths
        """
        return tuple(os.path.relpath(x, self._repo_dir) for x in paths)

    @property
    def actors(self):
        """
        :return: Tuple of actors in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.ACTOR, ()))

    @property
    def apis(self):
        """
        :return: Tuple of apis in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.API, ()))

    @property
    def topics(self):
        """
        :return: Tuple of topics in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.TOPIC, ()))

    @property
    def models(self):
        """
        :return: Tuple of models in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.MODEL, ()))

    @property
    def tags(self):
        """
        :return: Tuple of tags in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.TAG, ()))

    @property
    def workflows(self):
        """
        :return: Tuple of workflows in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.WORKFLOW, ()))

    @property
    def tools(self):
        """
        :return: Tuple of tools in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.TOOLS, ()))

    @property
    def libraries(self):
        """
        :return: Tuple of libraries in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.LIBRARIES, ()))

    @property
    def files(self):
        """
        :return: Tuple of files in the repository
        """
        return tuple(self._definitions.get(DefinitionKind.FILES, ()))



import json
import logging
import os
import sys

from leapp.exceptions import LeappError
from leapp.models import get_models
from leapp.repository.scan import find_and_scan_repositories
from leapp.snactor.utils import safe_discover
from leapp.tags import get_tags
from leapp.topics import get_topics
from leapp.utils.clicmd import command, command_opt
from leapp.utils.repository import find_repository_basedir, get_repository_name, requires_repository
from leapp.workflows import get_workflows


def _is_local(repository, cls, base_dir, all_repos=False):
    cls_path = os.path.realpath(sys.modules[cls.__module__].__file__)
    if all_repos:
        return any((cls_path.startswith(repo.repo_dir) for repo in repository.repos))
    return cls_path.startswith(base_dir)


def _get_class_file(cls, repository_relative=True):
    path = os.path.abspath(sys.modules[cls.__module__].__file__.replace('.pyc', '.py'))
    return os.path.relpath(path, find_repository_basedir('.') if repository_relative else os.getcwd())


def _print_group(name, items, name_resolver=lambda i: i.__name__, path_resolver=_get_class_file):
    sys.stdout.write('{group}({count}):\n'.format(group=name, count=len(items)))
    for item in sorted(items, key=name_resolver):
        sys.stdout.write('   - {name:<35} {path}\n'.format(name=name_resolver(item), path=path_resolver(item, False)))
    sys.stdout.write('\n')


def _get_actor_path(actor, repository_relative=True):
    path = actor.directory
    return os.path.relpath(path, find_repository_basedir('.') if repository_relative else os.getcwd())


def _get_actor_details(actor):
    meta = actor.discover()
    meta['produces'] = tuple(model.__name__ for model in actor.produces)
    meta['consumes'] = tuple(model.__name__ for model in actor.consumes)
    meta['apis'] = tuple(api.serialize() for api in actor.apis)
    meta['tags'] = tuple(tag.name for tag in meta['tags'])
    meta['path'] = _get_class_file(actor)
    meta['dialogs'] = [dialog.serialize() for dialog in actor.dialogs]
    return meta


def _get_workflow_details(workflow):
    return workflow.serialize()


def _get_tag_details(tag):
    return {'actors': [actor.class_name for actor in tag.actors],
            'name': tag.name}


def _get_topic_details(topic):
    return {'name': topic().name,
            'path': _get_class_file(topic)}


def _get_model_details(model):
    return {'path': _get_class_file(model)}


_LONG_DESCRIPTION = '''
Discovers and displays supported entities from the current repository.

Support entities:
- Actors
- Models
- Tags
- Topics
- Workflows


For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('discover', help="Discovers all available entities in the current repository",
         description=_LONG_DESCRIPTION)
@command_opt('json', is_flag=True, help='Output in json format instead of human readable form')
@command_opt('all', is_flag=True, help='Include items from linked repositories')
@command_opt('safe', is_flag=True, help='Analyze actor files statically to work around runtime errors')
@requires_repository
def cli(args):
    logging.basicConfig(level=logging.WARNING, stream=sys.stderr)
    base_dir = find_repository_basedir('.')

    if args.safe and args.json:
        sys.stderr.write('The options --safe and --json are currently mutually exclusive\n')
        sys.exit(1)

    if args.safe:
        sys.stdout.write(
            'Repository:\n  Name: {repository}\n  Path: {base_dir}\n\n'.format(repository=get_repository_name(base_dir),
                                                                               base_dir=base_dir))
        safe_discover(base_dir)
        sys.exit(0)

    repository = find_and_scan_repositories(base_dir, include_locals=True)
    try:
        repository.load()
    except LeappError as exc:
        sys.stderr.write(exc.message)
        sys.stderr.write('\n')
        sys.exit(1)

    actors = repository.actors
    topics = [topic for topic in get_topics() if _is_local(repository, topic, base_dir, all_repos=args.all)]
    models = [model for model in get_models() if _is_local(repository, model, base_dir, all_repos=args.all)]
    tags = [tag for tag in get_tags() if _is_local(repository, tag, base_dir, all_repos=args.all)]
    workflows = [workflow for workflow in get_workflows() if _is_local(repository, workflow, base_dir,
                                                                       all_repos=args.all)]
    if not args.json:
        sys.stdout.write(
            'Repository:\n  Name: {repository}\n  Path: {base_dir}\n\n'.format(repository=get_repository_name(base_dir),
                                                                               base_dir=base_dir))
        _print_group('Actors', actors, name_resolver=lambda x: x.class_name, path_resolver=_get_actor_path)
        _print_group('Models', models)
        _print_group('Tags', tags)
        _print_group('Topics', topics)
        _print_group('Workflows', workflows)
    else:
        output = {
            'repository': get_repository_name(base_dir),
            'base_dir': base_dir,
            'topics': dict((topic.__name__, _get_topic_details(topic)) for topic in topics),
            'models': dict((model.__name__, _get_model_details(model)) for model in models),
            'actors': dict((actor.class_name, _get_actor_details(actor)) for actor in actors),
            'tags': dict((tag.name, _get_tag_details(tag)) for tag in tags),
            'workflows': dict((workflow.__name__, _get_workflow_details(workflow)) for workflow in workflows)
        }
        json.dump(output, sys.stdout, indent=2)
        sys.stdout.write('\n')



import os
import pkgutil
import socket
import sys
import textwrap

from leapp import VERSION
from leapp.cli import commands
from leapp.exceptions import UnknownCommandError
from leapp.utils.clicmd import command


@command('')
def cli(args):  # noqa; pylint: disable=unused-argument
    """
    Top level base command dummy function
    """


def _load_commands(base_command):
    pkgdir = os.path.dirname(commands.__file__)
    for entry in os.listdir(pkgdir):
        entry_path = os.path.join(pkgdir, entry)
        if os.path.isdir(entry_path) and os.path.isfile(os.path.join(entry_path, '__init__.py')):
            # We found a package - We will import it and get the `register` symbol and check if it is callable.
            package_name = 'leapp.cli.commands.{}'.format(entry)
            package = pkgutil.get_loader(package_name).load_module(package_name)
            register = getattr(package, 'register', None)
            if callable(register):
                register(base_command)


def main():
    """
    leapp entry point
    """

    if os.getuid() != 0:
        sys.stderr.write('Leapp has to be executed with root privileges.\n')
        sys.exit(1)

    os.environ['LEAPP_HOSTNAME'] = socket.getfqdn()
    _load_commands(cli.command)
    try:
        cli.command.execute('leapp version {}'.format(VERSION))
    except UnknownCommandError as e:
        bad_cmd = (
            "Command \"{CMD}\" is unknown.\nMost likely there is a typo in the command or particular "
            "leapp repositories that provide this command are not present on the system.\n"
            "You can try to install the missing content e.g. by the following command: "
            "`dnf install 'leapp-command({CMD})'`")
        if e.requested.startswith('-'):
            # A quick ack not to confuse users with install a leapp-command(--some-wrong-argument) suggestion
            bad_cmd = "No such argument {CMD}"
        print(bad_cmd.format(CMD=e.requested))
        sys.exit(1)



from leapp.utils.audit import dict_factory, get_connection


def _fetch_table_for_context(db, table, context):
    cursor = db.execute('''
            SELECT * FROM {table} WHERE context = ?
        '''.format(table=table), (context,))
    cursor.row_factory = dict_factory
    while True:
        row = cursor.fetchone()
        if not row:
            break
        yield row
    del cursor


def _row_tuple(row, *fields):
    return tuple(row[name] for name in fields or row.keys())


def _dup_host(db, newcontext, oldcontext):
    lookup = {}
    for row in _fetch_table_for_context(db, 'host', oldcontext):
        # id, context, hostname
        row_id, hostname = _row_tuple(row, 'id', 'hostname')
        cursor = db.execute('INSERT INTO host (context, hostname) VALUES(?, ?)',
                            (newcontext, hostname))
        lookup[row_id] = cursor.lastrowid
    return lookup


def _dup_data_source(db, host, newcontext, oldcontext):
    lookup = {}
    for row in _fetch_table_for_context(db, 'data_source', oldcontext):
        # id, context, hostname
        row_id, host_id, actor, phase = _row_tuple(row, 'id', 'host_id', 'actor', 'phase')
        cursor = db.execute('INSERT INTO data_source (context, host_id, actor, phase) VALUES(?, ?, ?, ?)',
                            (newcontext, host[host_id], actor, phase))
        lookup[row_id] = cursor.lastrowid
    return lookup


def _dup_message(db, data_source, newcontext, oldcontext):
    lookup = {}
    for row in _fetch_table_for_context(db, 'message', oldcontext):
        # id, context, data_source_id, stamp, topic, type, message_data_hash
        row_id, data_source_id, stamp, topic, type_, message_data_hash = _row_tuple(
            row, 'id', 'data_source_id', 'stamp', 'topic', 'type', 'message_data_hash')
        cursor = db.execute(
            'INSERT INTO message (context, data_source_id, stamp, topic, type, message_data_hash) '
            ' VALUES(?, ?, ?, ?, ?, ?)',
            (newcontext, data_source[data_source_id], stamp, topic, type_, message_data_hash))
        lookup[row_id] = cursor.lastrowid
    return lookup


def _dup_audit(db, message, data_source, newcontext, oldcontext):
    lookup = {}
    for row in _fetch_table_for_context(db, 'audit', oldcontext):
        # id, context, event, stamp, data_source_id, message_id, data
        row_id, event, stamp, data_source_id, message_id, data = _row_tuple(
            row, 'id', 'event', 'stamp', 'data_source_id', 'message_id', 'data')
        if message_id is not None:
            message_id = message[message_id]

        cursor = db.execute(
            'INSERT INTO audit (context, event, stamp, data_source_id, message_id, data) VALUES(?, ?, ?, ?, ?, ?)',
            (newcontext, event, stamp, data_source[data_source_id], message_id, data))
        lookup[row_id] = cursor.lastrowid
    return lookup


def clone_context(oldcontext, newcontext, use_db=None):
    # Enter transaction - In case of any exception automatic rollback is issued
    # and it is automatically committed if there was no exception
    with get_connection(use_db) as db:
        # First clone host entries
        host = _dup_host(db=db, newcontext=newcontext, oldcontext=oldcontext)
        # Next clone data_source entries and use the lookup table generated by the host duplication
        data_source = _dup_data_source(db=db, host=host, newcontext=newcontext, oldcontext=oldcontext)
        # Next clone message entries and use the lookup table generated by the data_source duplication
        message = _dup_message(db=db, data_source=data_source, newcontext=newcontext, oldcontext=oldcontext)
        # Last clone message entries and use the lookup table generated by the data_source and message duplications
        _dup_audit(db=db, data_source=data_source, message=message, newcontext=newcontext, oldcontext=oldcontext)



from leapp.actors import Actor
from leapp.tags import FourthPhaseTag


class NotScheduledFourthActor(Actor):
    name = 'not_scheduled_fourth_actor'
    description = 'No description has been provided for the not_scheduled_fourth_actor actor.'
    consumes = ()
    produces = ()
    tags = (FourthPhaseTag,)

    def process(self):
        from leapp.libraries.common.test_helper import log_execution
        log_execution(self)



import json
import os
import sys
from importlib import import_module

from leapp.exceptions import CommandError, LeappError, ModelDefinitionError
from leapp.logger import configure_logger
from leapp.messaging.inprocess import InProcessMessaging
from leapp.models.fields import ModelViolationError
from leapp.repository.scan import find_and_scan_repositories
from leapp.snactor.context import with_snactor_context
from leapp.utils.audit import get_config, get_connection
from leapp.utils.clicmd import command, command_arg, command_opt
from leapp.utils.repository import find_repository_basedir, requires_repository

_MAIN_LONG_DESCRIPTION = '''
This group of commands are around managing messages stored in the
current repository database.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('messages', help='Messaging related commands', description=_MAIN_LONG_DESCRIPTION)
def messages(args):  # noqa; pylint: disable=unused-argument
    pass


_CLEAR_LONG_DESCRIPTION = '''
With this command messages in the current repository scope can be wiped.
This gives the developer control over the input data available to actors
run and developed in this repository.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


_ADD_LONG_DESCRIPTION = '''
With this command messages in the current repository scope can be added.
This gives the developer control over the input data available to actors
run and developed in this repository.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@messages.command('clear', help='Deletes all messages from the current repository scope',
                  description=_ADD_LONG_DESCRIPTION)
@requires_repository
@with_snactor_context
def clear(args):  # noqa; pylint: disable=unused-argument
    print("Deleting all messages with context = {} in database {}".format(
        os.environ["LEAPP_EXECUTION_ID"],
        get_config().get("database", "path")
    ))
    with get_connection(None) as con:
        con.execute("DELETE FROM message WHERE context = ?", (os.environ["LEAPP_EXECUTION_ID"],))


class Injector(object):
    name = 'injector'


@messages.command('add', help='Adds a message to the current repository scope',
                  description=_CLEAR_LONG_DESCRIPTION)
@command_opt('model', short_name='m', help='Model to add')
@command_arg('data', help='Data to add in JSON format')
@requires_repository
@with_snactor_context
def add(args):  # noqa; pylint: disable=unused-argument
    basedir = find_repository_basedir('.')
    repository = find_and_scan_repositories(basedir, include_locals=True)
    try:
        repository.load()
    except LeappError as exc:
        sys.stderr.write(exc.message)
        sys.stderr.write('\n')
        sys.exit(1)
    try:
        model = getattr(import_module('leapp.models'), args.model, None)
        InProcessMessaging().produce(model.create(json.loads(args.data)), Injector())
    except ModelDefinitionError:
        raise CommandError('No such model: {}'.format(args.model))
    except ModelViolationError as e:
        raise CommandError('Invalid data format for model: {}\n    => {}'.format(args.model, e.message))



from contextlib import contextmanager
import logging
import os
from subprocess import check_output

import py
import pytest

from leapp.actors import Actor, get_actors
from leapp.libraries.stdlib import api
from leapp.messaging import BaseMessaging
from leapp.models import ApiTestConsume, ApiTestProduce
from leapp.repository.scan import scan_repo


logging.basicConfig(format='%(asctime)-15s %(message)s')


class _TestableMessaging(BaseMessaging):
    def __init__(self, stored=True):
        super(_TestableMessaging, self).__init__(stored=stored)
        self.produced = []
        self.errors = []

    def produce(self, model, actor):
        self.produced.append(model)
        return super(_TestableMessaging, self).produce(model, actor)

    def _process_message(self, message):
        pass

    def _perform_load(self, consumes):
        pass

    def _do_produce(self, model, actor, target, stored=True):
        if type(model).__name__ == 'ErrorModel':
            self.errors.append(model)
        return super(_TestableMessaging, self)._do_produce(model=model, actor=actor, target=target, stored=stored)


@pytest.fixture(scope='module')
def repository(leapp_forked):  # noqa; pylint: disable=unused-argument
    repository_path = py.path.local(
        os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'actor-api-tests'))
    with repository_path.as_cwd():
        repo = scan_repo('.')
        repo.load(resolve=True)
        yield repo


@contextmanager
def _with_loaded_actor(repository, actor_name, messaging=None):
    definition = repository.lookup_actor(actor_name)
    with py.path.local(definition.full_path).as_cwd():
        # Ensure environmental cleanup after the test
        with definition.injected_context():
            # Load the actor class
            definition.load()
            # Lookup the actor class
            actor = [a for a in get_actors() if a.name == definition.name][0]
            yield (definition, actor(messaging=messaging, logger=logging.getLogger('root')))


@pytest.mark.parametrize("actor_name", ('first', 'second'))
def test_actor_api(repository, actor_name):
    with _with_loaded_actor(repository, actor_name) as (_unused, actor):
        # Ensure the current instance is exactly the current actor
        assert actor == Actor.current_instance
        assert actor == api.current_actor()
        assert api.current_logger() == actor.log


@pytest.mark.parametrize("actor_name", ('first', 'second'))
def test_actor_messaging_paths(leapp_forked, repository, actor_name):  # noqa; pylint: disable=unused-argument
    messaging = _TestableMessaging()
    with _with_loaded_actor(repository, actor_name, messaging) as (_unused, actor):
        messaging.feed(ApiTestConsume(data='prefilled'), actor)

        assert len(list(actor.consume(ApiTestConsume))) == 1
        assert next(actor.consume(ApiTestConsume)).data == 'prefilled'

        assert len(list(api.consume(ApiTestConsume))) == 1
        assert next(api.consume(ApiTestConsume)).data == 'prefilled'

        actor_message = 'Actor {} sent message via Actor'.format(actor_name)
        api_message = 'Actor {} sent message via API'.format(actor_name)

        actor.produce(ApiTestProduce(data=actor_message))
        assert messaging.produced.pop().data == actor_message

        api.produce(ApiTestProduce(data=api_message))
        assert messaging.produced.pop().data == api_message

        api.report_error("api error report", details={'source': 'api'})
        assert messaging.errors.pop().message.startswith("api ")
        actor.report_error("actor error report", details={'source': 'actor'})
        assert messaging.errors.pop().message.startswith("actor ")


@pytest.mark.parametrize("actor_name", ('first', 'second'))
def test_actor_all_files_paths(leapp_forked, repository, actor_name):  # noqa; pylint: disable=unused-argument
    with _with_loaded_actor(repository, actor_name) as (definition, actor):
        # API consistency verification
        assert api.files_paths() == actor.files_paths
        assert api.common_files_paths() == actor.common_files_paths
        assert api.actor_files_paths() == actor.actor_files_paths

        # Ensure environment and actor api results are the same
        assert ':'.join(actor.actor_files_paths) == os.getenv('LEAPP_FILES')
        assert ':'.join(actor.common_files_paths) == os.getenv('LEAPP_COMMON_FILES')
        assert ':'.join(actor.actor_tools_paths) == os.getenv('LEAPP_TOOLS')
        assert ':'.join(actor.common_tools_paths) == os.getenv('LEAPP_COMMON_TOOLS')

        # Here we must ensure that the sorted list of entries are correct
        assert sorted(actor.files_paths) == sorted(
            os.getenv('LEAPP_FILES').split(':') + os.getenv('LEAPP_COMMON_FILES').split(':'))

        assert sorted(actor.tools_paths) == sorted(
            os.getenv('LEAPP_TOOLS').split(':') + os.getenv('LEAPP_COMMON_TOOLS').split(':'))

        # Ensure LEAPP_FILES/actor_files_paths are really actor private
        for directory in actor.actor_files_paths:
            assert directory.startswith(definition.full_path)

        # Ensure the duplicate is resolvable from the actor private files and the right file
        assert actor.get_actor_file_path('duplicate')
        with open(actor.get_actor_file_path('duplicate')) as f:
            assert f.read().strip() == actor_name + '-actor'

        # Do the same thing with the API
        assert api.get_actor_file_path('duplicate')
        with open(api.get_actor_file_path('duplicate')) as f:
            assert f.read().strip() == actor_name + '-actor'

        # Ensure the duplicate is resolvable from the repository files and the right file
        assert actor.get_common_file_path('duplicate')
        with open(actor.get_common_file_path('duplicate')) as f:
            assert f.read().strip() == 'repository'

        # Do the same thing with the API
        assert api.get_common_file_path('duplicate')
        with open(api.get_common_file_path('duplicate')) as f:
            assert f.read().strip() == 'repository'

        # Ensure tools files api
        assert api.get_common_tool_path('directory/exec_script')
        assert check_output(api.get_common_tool_path('directory/exec_script')).decode('utf-8').strip() == 'repository'
        assert not api.get_common_tool_path('directory/nonexec_script')

        assert api.get_actor_tool_path('directory/exec_script')
        assert check_output(api.get_actor_tool_path('directory/exec_script'))\
            .decode('utf-8').strip() == actor_name + '-actor'
        assert not api.get_actor_tool_path('directory/nonexec_script')

        # Ensure some file is resolvable from the repository files or the actor private files
        assert actor.get_file_path('duplicate')
        assert actor.get_file_path('duplicate') == api.get_file_path('duplicate')

        check_path = 'directory/{}-actor'.format(actor_name)
        assert actor.get_file_path(check_path)
        assert api.get_file_path(check_path) == actor.get_file_path(check_path)

        assert actor.get_file_path('directory/repository')
        assert api.get_file_path('directory/repository') == actor.get_file_path('directory/repository')

        assert not actor.get_common_file_path('directory/{}-actor'.format(actor_name))
        assert not api.get_common_file_path('directory/{}-actor'.format(actor_name))

        assert not actor.get_actor_file_path('directory/repository')
        assert not api.get_actor_file_path('directory/repository')



import json
from datetime import datetime

import pytest
import six

from leapp.models import Model, fields
from leapp.topics import Topic


class BadBuiltinField(fields.BuiltinField):  # noqa; pylint: disable=abstract-method
    pass


class ModelTestTopic(Topic):
    name = 'model-test-topic'


class BasicModel(Model):
    topic = ModelTestTopic
    message = fields.String(default='Default Value')


class WithStringListModel(Model):
    topic = ModelTestTopic
    messages = fields.List(fields.String())


class WithNestedModel(Model):
    topic = ModelTestTopic
    basic = fields.Model(BasicModel)


class WithNullableNestedModel(Model):
    topic = ModelTestTopic
    basic = fields.Nullable(fields.Model(BasicModel))


class WithNestedListModel(Model):
    topic = ModelTestTopic
    items = fields.List(fields.Model(BasicModel))


class WithBlobModel(Model):
    topic = ModelTestTopic
    message = fields.Blob()
    can_be_empty = fields.Nullable(fields.Blob())


class AllFieldTypesModel(Model):
    topic = ModelTestTopic
    float_field = fields.Float(default=3.14)
    number_int_field = fields.Number(default=1.2)
    number_float_field = fields.Number(default=2)
    integer_field = fields.Integer(default=1)
    str_field = fields.String(default='string')
    unicode_field = fields.String(default=u'Unicode string')
    date_field = fields.DateTime(default=datetime.utcnow())
    bool_field = fields.Boolean(default=True)


class RequiredFieldModel(Model):
    topic = ModelTestTopic
    field = fields.String()


def test_builtin_needs_override():
    with pytest.raises(NotImplementedError):
        fields.Nullable(BadBuiltinField()).to_builtin(None, '', None)


def test_base_usage():
    with pytest.raises(fields.ModelMisuseError):
        fields.Field()


def test_basic_model():
    m = BasicModel(message='Some message')
    m2 = BasicModel.create(m.dump())
    assert m.message == m2.message


def test_bytestring_model():
    m = WithBlobModel(message=b'\xf3\xcf\xcf\xc2\xdd\xc5\xce\xc9\xc5')
    m2 = WithBlobModel.create(m.dump())
    assert m.message == m2.message


def test_string_list_model():
    m = WithStringListModel(messages=['Some message'])
    m2 = WithStringListModel.create(m.dump())
    assert m.messages == m2.messages
    m2.messages = 'str'

    with pytest.raises(fields.ModelViolationError):
        m2.dump()

    with pytest.raises(fields.ModelViolationError):
        WithStringListModel(messages='str')


def test_string_fields_violations():
    f = fields.String()
    with pytest.raises(fields.ModelViolationError):
        f._validate_model_value(1, 'test_value')

    with pytest.raises(fields.ModelViolationError):
        f._validate_builtin_value(1, 'test_value')


def test_nested_model():
    m = WithNestedModel(basic=BasicModel(message='Some message'))
    m2 = WithNestedModel.create(m.dump())
    assert m.basic == m2.basic

    with pytest.raises(fields.ModelMisuseError):
        fields.Model(fields.String())

    with pytest.raises(fields.ModelMisuseError):
        fields.Model(fields.String)

    with pytest.raises(fields.ModelViolationError):
        WithNestedModel(basic='Some message')

    m = WithNullableNestedModel()
    m.basic = None
    m.dump()

    with pytest.raises(fields.ModelViolationError):
        x = WithNestedModel(basic=BasicModel(message='Some message'))
        x.basic = None
        x.dump()

    with pytest.raises(fields.ModelViolationError):
        WithNestedModel.create(dict(basic=None))

    with pytest.raises(fields.ModelViolationError):
        WithNestedModel(basic=None)

    assert WithNestedModel.create({'basic': {'message': 'test-message'}}).basic.message == 'test-message'
    assert WithNestedModel(basic=BasicModel(message='test-message')).basic.message == 'test-message'


def test_nested_list_model():
    m = WithNestedListModel(items=[BasicModel(message='Some message')])
    m2 = WithNestedListModel.create(m.dump())
    assert m.items == m2.items


def test_field_types():
    m = AllFieldTypesModel()
    m2 = AllFieldTypesModel.create(m.dump())
    assert m == m2


def test_misuse_wrong_list_element_parameter():
    with pytest.raises(fields.ModelMisuseError):
        class RaisesNonFieldType(Model):
            topic = ModelTestTopic
            boo = fields.List('')

    with pytest.raises(fields.ModelMisuseError):
        class RaisesNonFieldInstance(Model):
            topic = ModelTestTopic
            boo = fields.List(str)

    with pytest.raises(fields.ModelMisuseError):
        class RaisesNonInstance(Model):
            topic = ModelTestTopic
            boo = fields.List(fields.String)


def test_list_field():
    with pytest.raises(fields.ModelViolationError):
        fields.List(fields.String())._validate_builtin_value('something', 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.List(fields.String())._convert_to_model(None, 'test-value')

    fields.Nullable(fields.List(fields.String()))._convert_to_model(None, 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.List(fields.String())._convert_from_model(None, 'test-value')

    fields.Nullable(fields.List(fields.String()))._convert_from_model(None, 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.Nullable(fields.List(fields.Integer(), minimum=1))._validate_builtin_value([], 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.Nullable(fields.List(fields.Integer(), minimum=1))._validate_model_value([], 'test-value')

    fields.List(fields.Integer(), minimum=1)._validate_builtin_value([1], 'test-value')
    fields.List(fields.Integer(), minimum=1)._validate_builtin_value([1, 2], 'test-value')
    fields.List(fields.Integer(), minimum=1)._validate_model_value([1], 'test-value')
    fields.List(fields.Integer(), minimum=1)._validate_model_value([1, 2], 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.List(fields.Integer(), minimum=1, maximum=1)._validate_builtin_value([1, 2], 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.List(fields.Integer(), minimum=1, maximum=1)._validate_model_value([1, 2], 'test-value')

    fields.List(fields.Integer(), maximum=2)._validate_builtin_value([1], 'test-value')
    fields.List(fields.Integer(), maximum=2)._validate_builtin_value([1, 2], 'test-value')

    fields.List(fields.Integer(), maximum=2)._validate_model_value([1], 'test-value')
    fields.List(fields.Integer(), maximum=2)._validate_model_value([1, 2], 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.List(fields.Integer(), maximum=3)._validate_builtin_value([1, 2, 3, 4], 'test-value')


def test_datetime_field():
    with pytest.raises(fields.ModelViolationError):
        fields.DateTime()._convert_to_model('something', 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.DateTime()._convert_to_model(None, 'test-value')

    fields.Nullable(fields.DateTime())._convert_to_model(None, 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.DateTime()._convert_from_model(None, 'test-value')

    fields.Nullable(fields.DateTime())._convert_from_model(None, 'test-value')


def test_nested_field():
    with pytest.raises(fields.ModelViolationError):
        fields.Model(BasicModel)._convert_to_model('something', 'test-value')
    with pytest.raises(fields.ModelViolationError):
        fields.Model(BasicModel)._convert_to_model(None, 'test-value')
    fields.Nullable(fields.Model(BasicModel))._convert_to_model(None, 'test-value')


def test_required_field_types():
    # all fields which are not nullable are required
    with pytest.raises(fields.ModelViolationError):
        m = RequiredFieldModel(field='str')
        m.field = None
        m.dump()

    with pytest.raises(fields.ModelViolationError):
        RequiredFieldModel()

    RequiredFieldModel(field='str')

    with pytest.raises(fields.ModelViolationError):
        fields.String()._validate_model_value(None, 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.String()._validate_model_value(None, 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.String()._validate_builtin_value(None, 'test-value')

    with pytest.raises(fields.ModelViolationError):
        fields.String()._validate_builtin_value(None, 'test-value')


def test_not_required_field_types():
    fields.Nullable(fields.String())._validate_model_value(None, 'test-value')
    fields.Nullable(fields.String())._validate_model_value(None, 'test-value')
    fields.Nullable(fields.String())._validate_builtin_value(None, 'test-value')
    fields.Nullable(fields.String())._validate_builtin_value(None, 'test-value')


def _make_object(instance, value):
    def init(self):
        setattr(self, instance.name, value)

    return type('Dynamic' + instance.field_type.__name__, (object,), {'__init__': init})()


def _make_dict(instance, value):
    return {instance.name: value}


def create_fixture(field, value, name, fail_values=None):
    def init(self):
        setattr(self, 'field_type', field)

    return type('Fixture', (object,), {
        'value': value,
        'name': name,
        'fail_values': fail_values,
        'make_object': _make_object,
        'make_dict': _make_dict,
        '__init__': init
    })()


def _create_field_string_list(**kwargs):
    return fields.List(fields.String(), **kwargs)


def _create_nested_base_model_field(**kwargs):
    return fields.Model(BasicModel, **kwargs)


def _create_string_enum_(*choices):
    def _fun(**kwargs):
        return fields.StringEnum(choices=choices, **kwargs)
    return _fun


BASIC_TYPE_FIXTURES = (
    create_fixture(fields.String, 'Test String Value', 'string_value'),
    create_fixture(fields.Boolean, True, 'boolean_value'),
    create_fixture(fields.Integer, 1, 'integer_value'),
    create_fixture(fields.Float, 3.14, 'float_value'),
    create_fixture(fields.Number, 3.14, 'number_float_value'),
    create_fixture(fields.Number, 2, 'number_integer_value'),
    create_fixture(fields.DateTime, datetime.utcnow(), 'datetime_value'),
    create_fixture(_create_field_string_list, ['a', 'b', 'c'], 'string_list_value'),
    create_fixture(_create_nested_base_model_field, BasicModel(message='Test message'), 'nested_model_value'),
    create_fixture(_create_string_enum_('YES', 'NO', 'MAYBE'), 'YES', 'string_enum_value', fail_values=('Woot',)),
    create_fixture(fields.JSON, 2, 'json_integral_value'),
    create_fixture(fields.JSON, 3.14, 'json_float_value'),
    create_fixture(fields.JSON, 'some string', 'json_string_value'),
    create_fixture(fields.JSON, ['a', 'b', 1, 2], 'json_list_value'),
    create_fixture(fields.JSON, ('a', 'b', 1, 2), 'json_tuple_value'),
    create_fixture(fields.JSON, {'a': [None, {'b': 'c'}], 'b': 2}, 'json_dict_value'),
    create_fixture(fields.JSON, True, 'json_bool_value'),
)


def test_choices_wrong_value():
    with pytest.raises(fields.ModelMisuseError):
        fields.StringEnum(choices='abc')
    fields.StringEnum(choices=('abc',))
    with pytest.raises(fields.ModelMisuseError):
        fields.IntegerEnum(choices=1)
    fields.IntegerEnum(choices=(1,))
    with pytest.raises(fields.ModelMisuseError):
        fields.FloatEnum(choices=1.2)
    fields.FloatEnum(choices=(1.2,))
    with pytest.raises(fields.ModelMisuseError):
        fields.NumberEnum(choices=3.14)
    fields.NumberEnum(choices=(3.14,))


def test_list_field_default():
    fields.List(fields.StringEnum(choices=('1', '2', '3')), default=['1', '2'])


@pytest.mark.parametrize("case", BASIC_TYPE_FIXTURES)
def test_basic_types_sanity(case):
    source = case.make_object(case.value)
    target = {}
    case.field_type().to_builtin(source, case.name, target)
    json.dumps(target)

    source = case.make_object(None)
    target = {}
    case.field_type().as_nullable().to_builtin(source, case.name, target)
    assert case.name in target and target[case.name] is None

    with pytest.raises(fields.ModelViolationError):
        source = case.make_object(None)
        target = {}
        # Should raise an exception because the field is required and null is not allowed
        case.field_type().to_builtin(source, case.name, target)

    with pytest.raises(fields.ModelViolationError):
        source = case.make_object(None)
        target = {}
        # Should raise an exception because the field null is not allowed
        case.field_type().to_builtin(source, case.name, target)

    source = case.make_object(case.value)
    target = {}
    field = case.field_type()
    field.to_builtin(source, case.name, target)
    assert target.get(case.name) == field._convert_from_model(case.value, case.name)
    json.dumps(target)

    source = case.make_object(None)
    target = {}
    field = case.field_type().as_nullable()
    field.to_builtin(source, case.name, target)
    assert target.get(case.name) is None
    json.dumps(target)

    if case.fail_values:
        for fail_value in case.fail_values:
            source = case.make_object(fail_value)
            target = {case.name: fail_value}
            field = case.field_type()
            with pytest.raises(fields.ModelViolationError):
                field.to_builtin(source, case.name, target)

            target = case.make_object(case.value)
            source = {case.name: fail_value}
            with pytest.raises(fields.ModelViolationError):
                field.to_model(source, case.name, target)

    assert isinstance(field.help, six.string_types)



import pytest

from leapp.exceptions import InvalidTagDefinitionError
from leapp.tags import Tag, get_tags, ExperimentalTag, DisabledTag


class TestTag(Tag):
    name = "test-tag"


def test_tag_members_correctly_set():
    assert hasattr(TestTag, 'Common') and TestTag.Common.name == 'common-test-tag'
    assert hasattr(TestTag, 'Before') and TestTag.Before.name == 'before-test-tag'
    assert hasattr(TestTag.Before, 'Common') and TestTag.Before.Common.name == 'common-before-test-tag'
    assert hasattr(TestTag, 'After') and TestTag.After.name == 'after-test-tag'
    assert hasattr(TestTag.After, 'Common') and TestTag.After.Common.name == 'common-after-test-tag'
    assert not hasattr(TestTag.Common, 'Common')
    assert not hasattr(TestTag.Common, 'After')
    assert not hasattr(TestTag.Common, 'Before')
    assert not hasattr(TestTag.After, 'Before')
    assert not hasattr(TestTag.After, 'After')
    assert not hasattr(TestTag.Before, 'Before')
    assert not hasattr(TestTag.Before, 'After')
    assert not TestTag.actors
    assert not TestTag.Common.actors
    assert not TestTag.Before.actors
    assert not TestTag.After.actors
    assert not TestTag.Before.Common.actors
    assert not TestTag.After.Common.actors
    assert not hasattr(TestTag, 'parent')
    assert TestTag.Common.parent is TestTag
    assert TestTag.Before.parent is TestTag
    assert TestTag.After.parent is TestTag
    assert TestTag.Before.Common.parent is TestTag
    assert TestTag.After.Common.parent is TestTag


def test_get_tags():
    tags = [tag for tag in get_tags() if tag not in (ExperimentalTag, DisabledTag)]
    assert len(tags) % 6 == 0
    assert TestTag in tags
    assert TestTag.Common in tags
    assert TestTag.Before in tags
    assert TestTag.After in tags
    assert TestTag.Before.Common in tags
    assert TestTag.After.Common in tags
    TestTag.name = None
    with pytest.raises(InvalidTagDefinitionError):
        get_tags()



import datetime
import logging
import logging.config
import os
import sys
import time

from leapp.config import get_config
from leapp.libraries.stdlib.config import is_debug, is_verbose
from leapp.utils.actorapi import get_actor_api, RequestException
from leapp.utils.audit import Audit

_logger = None


class LeappAuditHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super(LeappAuditHandler, self).__init__(*args, **kwargs)
        self.use_remote = kwargs.pop('use_remote', False)
        if self.use_remote:
            self.url = 'leapp://localhost/actors/v1/log'
            self.session = get_actor_api()

    def emit(self, record):
        log_data = {
            'event': 'log-message',
            'context': os.environ.get('LEAPP_EXECUTION_ID', 'TESTING-CONTEXT'),
            'stamp': datetime.datetime.utcnow().isoformat() + 'Z',
            'hostname': os.environ.get('LEAPP_HOSTNAME', 'localhost'),
            'actor': os.environ.get('LEAPP_CURRENT_ACTOR', ''),
            'phase': os.environ.get('LEAPP_CURRENT_PHASE', ''),
            'log': {
                'level': record.levelname,
                'message': self.format(record)
            }
        }
        if self.use_remote:
            self._remote_emit(log_data)
        else:
            self._do_emit(log_data)

    @staticmethod
    def _do_emit(log_data):
        log_data['data'] = log_data.pop('log', {})
        Audit(**log_data).store()

    def _remote_emit(self, log_data):
        try:
            self.session.post(self.url, json=log_data, timeout=0.1)
        except RequestException:
            pass


def configure_logger(log_file=None):
    global _logger
    if not _logger:
        log_format = '%(asctime)s.%(msecs)-3d %(levelname)-8s PID: %(process)d %(name)s: %(message)s'
        log_date_format = '%Y-%m-%d %H:%M:%S'
        path = os.getenv('LEAPP_LOGGER_CONFIG', '/etc/leapp/logger.conf')

        if path and os.path.isfile(path):
            logging.config.fileConfig(path, disable_existing_loggers=False)
        else:  # Fall back logging configuration
            logging.Formatter.converter = time.gmtime
            logging.basicConfig(
                level=logging.ERROR,
                format=log_format,
                datefmt=log_date_format,
                stream=sys.stderr,
            )
            logging.getLogger('urllib3').setLevel(logging.WARN)
            handler = LeappAuditHandler()
            handler.setFormatter(logging.Formatter(fmt=log_format, datefmt=log_date_format))
            logging.getLogger('leapp').addHandler(handler)

        if log_file:
            file_handler = logging.FileHandler(os.path.join(get_config().get('logs', 'dir'), log_file))
            file_handler.setFormatter(logging.Formatter(fmt=log_format, datefmt=log_date_format))
            file_handler.setLevel(logging.DEBUG)
            logging.getLogger('leapp').addHandler(file_handler)

        if is_verbose():
            for handler in logging.getLogger().handlers:
                if isinstance(handler, logging.StreamHandler):
                    handler.setLevel(logging.DEBUG if is_debug() else logging.INFO)

        _logger = logging.getLogger('leapp')
        _logger.info('Logging has been initialized')

    return _logger



import os

import sys

from leapp.utils.clicmd import command_arg, command
from leapp.exceptions import CommandError
from leapp.utils.repository import requires_repository, make_class_name, make_name, find_repository_basedir

_LONG_DESCRIPTION = '''
Creates a new Tag in the current repository.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('new-tag', help='Create a new tag', description=_LONG_DESCRIPTION)
@command_arg('tag-name')
@requires_repository
def cli(args):
    basedir = os.path.join(find_repository_basedir('.'), 'tags')
    if not os.path.isdir(basedir):
        os.mkdir(basedir)

    tag_path = os.path.join(basedir, args.tag_name.lower() + '.py')
    if os.path.exists(tag_path):
        raise CommandError("File already exists: {}".format(tag_path))

    tag_class_name = make_class_name(args.tag_name)
    if not tag_class_name.endswith('Tag'):
        tag_class_name += 'Tag'
    with open(tag_path, 'w') as f:
        f.write('''from leapp.tags import Tag


class {tag_name}(Tag):
    name = '{tag}'
'''.format(tag_name=tag_class_name, tag=make_name(args.tag_name)))

    sys.stdout.write("New tag {} has been created in {}\n".format(tag_class_name, os.path.realpath(tag_path)))



from leapp.actors import Actor
from leapp.tags import InplaceUpgradeWorkflowTag, FirstBootTag


class FirstBootActor(Actor):
    """
    No documentation has been provided for the first_boot_actor actor.
    """

    name = 'first_boot_actor'
    consumes = ()
    produces = ()
    tags = (InplaceUpgradeWorkflowTag, FirstBootTag)

    def process(self):
        print('<<<TEST>>>: {}'.format(self.__class__.__name__))



"""
Functions that allow gather information about the current configuration.
"""
import os


def is_debug():
    """
    Returns True if the LEAPP_DEBUG environment variable is set to '1'.
    """
    return os.getenv('LEAPP_DEBUG', '0') == '1'


def is_verbose():
    """
    Returns True if either the LEAPP_DEBUG or the LEAPP_VERBOSE environment variable is set to '1'.
    """
    return os.getenv('LEAPP_DEBUG', '0') == '1' or os.getenv('LEAPP_VERBOSE', '0') == '1'



import os
import socket

import requests
import requests.adapters
import requests.exceptions

try:
    import requests.packages.urllib3 as urllib3  # noqa # pylint: disable=consider-using-from-import
except ImportError:
    import urllib3


RequestException = requests.exceptions.RequestException
"""
Exceptions that are raised through the actor API, which is retrieved from :py:func:`get_actor_api`
"""
_session = None


class _LeappAPIConnection(urllib3.connection.HTTPConnection):
    def __init__(self, timeout=60):
        self.sock = None
        super(_LeappAPIConnection, self).__init__('localhost', timeout=timeout)
        self.timeout = timeout

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        socket_path = os.environ.get('LEAPP_ACTOR_API', '/tmp/actors.sock')
        sock.connect(socket_path)
        self.sock = sock


class _LeappAPIConnectionPool(urllib3.connectionpool.HTTPConnectionPool):
    def __init__(self, timeout=60):
        super(_LeappAPIConnectionPool, self).__init__('localhost')
        self.timeout = timeout

    def _new_conn(self):
        return _LeappAPIConnection(self.timeout)


class _LeappAPIAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, timeout=60):
        super(_LeappAPIAdapter, self).__init__()
        self.timeout = timeout

    def get_connection(self, url, proxies=None):
        return _LeappAPIConnectionPool(timeout=self.timeout)

    def request_url(self, request, proxies):
        return request.path_url


def get_actor_api():
    """
    :return: An instance of the Leapp actor API session that is using :py:class:`requests.Session` over a
             UNIX domain socket
    """
    global _session
    if _session:
        return _session
    _session = requests.session()
    _session.adapters.clear()
    _session.mount('leapp://', _LeappAPIAdapter())
    return _session



RAISE_EXC_HELPER = '''
def raise_with_traceback(exc, tb):
    """
    This is a helper function to raise exceptions with a traceback.

    This is function is required to workaround the syntax changes between Python 2 and 3
    Python 3.4 introduced a with_traceback method to Exception classes and Python 3 removed the syntax
    which used to be used in Python 2.

    :param exc: Exception to raise
    :param tb:  Traceback to use
    :return: Nothing
    """
    raise exc, None, tb
    '''

exec(RAISE_EXC_HELPER, globals(), locals())  # noqa; pylint: disable=exec-used



class DefinitionKind(object):
    """
    Represents all known repository resources in Leapp.
    """
    class _Kind(object):
        def __init__(self, kind):
            self.name = kind

    ACTOR = _Kind('actor')
    MODEL = _Kind('model')
    TOPIC = _Kind('topic')
    TAG = _Kind('tag')
    WORKFLOW = _Kind('workflow')
    LIBRARIES = _Kind('libraries')
    TOOLS = _Kind('tools')
    FILES = _Kind('files')
    TESTS = _Kind('tests')
    API = _Kind('api')

    REPO_WHITELIST = (ACTOR, API, MODEL, TOPIC, TAG, WORKFLOW, TOOLS, LIBRARIES, FILES)
    ACTOR_WHITELIST = (TOOLS, LIBRARIES, FILES, TESTS)



import os

from leapp.actors import Actor
from leapp.dialogs import Dialog
from leapp.dialogs.components import BooleanComponent
from leapp.tags import FirstPhaseTag, UnitTestWorkflowTag


class FirstActor(Actor):
    name = 'first_actor'
    description = 'No description has been provided for the first_actor actor.'
    consumes = ()
    produces = ()
    tags = (FirstPhaseTag, UnitTestWorkflowTag)
    dialogs = (
        Dialog(
            scope='unique_dialog_scope',
            reason='Confirmation',
            components=(
                BooleanComponent(
                    key='confirm',
                    label='Disable a deprecated module?'
                          'If no, the upgrade process will be interrupted.',
                    default=False,
                    description='Module XXX is no longer available '
                                'in RHEL-8 since it was replaced by shiny-metal-XXX.',
                    reason='Leaving this module in system configuration may lock out the system.'
                ),
            )
        ),)

    def process(self):
        from leapp.libraries.common.test_helper import log_execution
        log_execution(self)
        if not self.configuration or self.configuration.value != 'unit-test':
            self.report_error('Unit test failed due missing or invalid workflow provided configuration')
        if os.environ.get('FirstActor-ReportError') == '1':
            self.report_error("Unit test requested error")
        self.get_answers(self.dialogs[0]).get('confirm', False)



from leapp.workflows.api.v1.depcheck import DepCheckAPI1, DepCheckAPI2, DepCheckAPI3, DepCheckAPI4
from leapp.workflows.api.v2.newapi import NewAPI
from leapp.workflows.api.v2.testapi import TestAPI


__all__ = (
    DepCheckAPI1,
    DepCheckAPI2,
    DepCheckAPI3,
    DepCheckAPI4,
    TestAPI,
    NewAPI
)



class LeappError(Exception):
    def __init__(self, message):
        super(LeappError, self).__init__(message)
        self.message = message


class RepositoryConfigurationError(LeappError):
    pass


class CannotConsumeErrorMessages(LeappError):
    def __init__(self):
        super(CannotConsumeErrorMessages, self).__init__("Actors cannot consume error messages.")


class InvalidTopicItemError(LeappError):
    pass


class InvalidTopicDefinitionError(LeappError):
    pass


class InvalidTagDefinitionError(LeappError):
    pass


class MissingActorAttributeError(LeappError):
    pass


class WrongAttributeTypeError(LeappError):
    pass


class ModelDefinitionError(LeappError):
    pass


class TagFilterUsageError(LeappError):
    pass


class CyclingDependenciesError(LeappError):
    pass


class UnsupportedDefinitionKindError(LeappError):
    pass


class ModuleNameAlreadyExistsError(LeappError):
    pass


class ActorInspectionFailedError(LeappError):
    pass


class MultipleActorsError(LeappError):
    def __init__(self, path):
        super(MultipleActorsError, self).__init__(
            'Multiple actors found in {path}. Inspection failed'.format(path=path))


class MultipleConfigActorsError(LeappError):
    def __init__(self, config_actors):
        super(MultipleConfigActorsError, self).__init__(
            'Multiple config actors detected: {config_actors}. '
            'Only one config actor per workflow is allowed'.format(config_actors=config_actors))


class WorkflowConfigNotAvailable(LeappError):
    def __init__(self, actor):
        # TODO(mreznik): Current implementation of the workflow congiguration is problematic when used
        # with snactor. See https://github.com/oamg/leapp/issues/530
        super(WorkflowConfigNotAvailable, self).__init__(
            'Actor {actor} relies on workflow configuration model which '
            'must be produced by a specific actor'.format(actor=actor))


class RepoItemPathDoesNotExistError(LeappError):
    def __init__(self, kind, rel_path, full_path):
        super(RepoItemPathDoesNotExistError, self).__init__(
            'Could not find {kind} item with relative path: {rel_path} at {full_path}'.format(
                kind=kind, rel_path=rel_path, full_path=full_path))


class ActorDiscoveryExecutionError(LeappError):
    pass


class UsageError(LeappError):
    pass


class CommandError(LeappError):
    pass


class CommandDefinitionError(LeappError):
    pass


class UnknownCommandError(LeappError):
    def __init__(self, command):
        super().__init__('Unknown command: {}'.format(command))
        self.requested = command


class LeappRuntimeError(LeappError):
    pass


class StopActorExecution(Exception):
    """ This exception is used to gracefully stop execution of actor, but allows the workflow to continue. """


class StopActorExecutionError(LeappError):
    """
    This exception is used to gracefully stop execution of actor and it will call
    :py:func:`leapp.actors.Actor.report_error`.
    """
    # import here to break import cycle
    from leapp.models.error_severity import ErrorSeverity  # pylint: disable=import-outside-toplevel

    def __init__(self, message, severity=ErrorSeverity.ERROR, details=None):
        """
        :param message: A message to print the possible error
        :type message: str
        :param severity: Severity of the error default :py:attr:`leapp.messaging.errors.ErrorSeverity.ERROR`
        :type severity: str with defined values from :py:attr:`leapp.messaging.errors.ErrorSeverity.ERROR`
        :param details: A dictionary where additional context information is passed along with the error
        :type details: dict
        """
        super(StopActorExecutionError, self).__init__(message)
        self.severity = severity
        self.details = details


class RequestStopAfterPhase(LeappError):
    """
    This exception is used to gracefully stop the current actor and request the stop of the workflow execution after
    the current phase.
    """

    def __init__(self):
        super(RequestStopAfterPhase, self).__init__('Stop after phase has been requested.')



import os
import multiprocessing.util
import multiprocessing

_start_method_set = False


def apply_workaround():
    global _start_method_set
    # Set start method for multiprocessing to fork (3.4+)
    set_start_method = getattr(multiprocessing, 'set_start_method', None)
    if not _start_method_set and set_start_method:
        set_start_method('fork')
        _start_method_set = True

    # Implements:
    # https://github.com/python/cpython/commit/e8a57b98ec8f2b161d4ad68ecc1433c9e3caad57
    #
    # Detection of fix: os imported to compare pids, before the fix os has not
    # been imported
    if getattr(multiprocessing.util, 'os', None):
        return

    class FixedFinalize(multiprocessing.util.Finalize):
        def __init__(self, *args, **kwargs):
            super(FixedFinalize, self).__init__(*args, **kwargs)
            self._pid = os.getpid()

        def __call__(self, *args, **kwargs):  # pylint: disable=signature-differs
            if self._pid != os.getpid():
                return None
            return super(FixedFinalize, self).__call__(*args, **kwargs)

    setattr(multiprocessing.util, 'Finalize', FixedFinalize)



from leapp.models import Model, fields
from leapp.topics import DeprecationTopic
from leapp.utils.deprecation import deprecated


@deprecated('2010-01-01', 'This model is deprecated - Please do not use it anymore')
class DeprecatedModel(Model):
    topic = DeprecationTopic


@deprecated('2010-01-01', 'This model is deprecated - This should not be warned about')
class SuppressedDeprecatedModel(Model):
    topic = DeprecationTopic



import sys

from leapp.exceptions import InvalidTagDefinitionError
from leapp.utils.meta import get_flattened_subclasses, with_metaclass


class TagMeta(type):
    """
    Meta class for the registration of tags

    This meta class adds dynamically Common, Before, Before.Common, After and After.Common attributes to the
    tag class. For more information see :py:class:`leapp.tags.Tag`
    """
    def __new__(mcs, name, bases, attrs):
        klass = super(TagMeta, mcs).__new__(mcs, name, bases, attrs)
        if klass.__module__ is not TagMeta.__module__:
            setattr(sys.modules[mcs.__module__], name, klass)
            setattr(klass, 'actors', ())

            if not getattr(klass, 'parent', None):
                data = {'parent': klass, 'actors': ()}
                before_common = type('_' + name + 'BeforeCommon', (Tag,), dict(name='common-before-' + klass.name,
                                                                               **data))
                after_common = type('_' + name + 'AfterCommon', (Tag,), dict(name='common-after-' + klass.name, **data))
                klass.Before = type('_' + 'Before' + name, (Tag,),
                                    dict(name='before-' + klass.name, Common=before_common, **data))
                klass.After = type('_' + 'After' + name, (Tag,),
                                   dict(name='after-' + klass.name, Common=after_common, **data))
                klass.Common = type('_' + 'Common' + name, (Tag,), dict(name='common-' + klass.name, **data))

                # To allow pickle to handle these tags we have to publish them on module level
                globals().update({
                    before_common.__name__: before_common,
                    after_common.__name__: after_common,
                    klass.Before.__name__: klass.Before,
                    klass.After.__name__: klass.After,
                    klass.Common.__name__: klass.Common})
        return klass


class Tag(with_metaclass(TagMeta)):
    """
    Tag is the base class for all Tags. Tags are used as filtering mechanism for actors to be loaded during workflow
    executions. Phases do use tags to filter actors according to their tags.

    Special Tag class attributes:
        Tag here refers to the derived class of Tag

        Tag.Common:
            Dynamically created class type that designates actors to be executed in the `main` stage during
            workflow phases. Using common includes this actor in any workflow, which means the that any workflow
            tag filter will be ignored if this tag matches.

        Tag.Before:
            Dynamically created class type that designates actors to be executed in the `before` stage during
            workflow phases.

        Tag.Before.Common:
            Dynamically created class type that designates actors to be executed in the `before` stage during
            workflow phases. Using common includes this actor in any workflow, which means the that any workflow
            tag filter will be ignored if this tag matches.

        Tag.After:
            Dynamically created class type that designates actors to be executed in the `after` stage during
            workflow phases.

        Tag.After.Common:
            Dynamically created class type that designates actors to be executed in the `after` stage during
            workflow phases. Using common includes this actor in any workflow, which means the that any workflow
            tag filter will be ignored if this tag matches.
    """

    actors = ()
    """
    Tuple of all registered actors using this tag
    """

    name = None
    """ Name of the Tag in snake case """

    @classmethod
    def serialize(cls):
        return {
            'class_name': cls.__name__,
            'name': cls.name,
            'actors': [actor.class_name for actor in cls.actors]
        }


class ExperimentalTag(Tag):
    name = 'experimental'


class DisabledTag(Tag):
    name = 'disabled'


def get_tags():
    """
    :return: All registered :py:class:`leapp.tags.Tag` derived classes
    """
    tags = get_flattened_subclasses(Tag)
    for tag in (tag for tag in tags if tag is not Tag):
        tag_name = getattr(tag, 'name', None)
        if not tag_name:
            raise InvalidTagDefinitionError('Tag {} does not contain a tag name attribute'.format(tag))
    return tags



from __future__ import print_function
import json
import os
import subprocess
import sys

from leapp.utils.clicmd import command, command_opt, command_arg
from leapp.utils.repository import requires_repository, find_repository_basedir, get_repository_name, \
    get_repository_id, add_repository_link, get_user_config_repos, get_user_config_repo_data, \
    get_global_repositories_data
from leapp.exceptions import CommandError, UsageError

_MAIN_LONG_DESCRIPTION = '''
This group of commands are around managing repositories.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('repo', help='Repository related commands', description=_MAIN_LONG_DESCRIPTION)
def repo(args):  # noqa; pylint: disable=unused-argument
    pass


_HEALTH_CHECK_LONG_DESCRIPTION = '''
Health check is used to remove stale repository entries from the user local repository registration.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@repo.command('health-check', help='Checks registered repositories and removes missing entries',
              description=_HEALTH_CHECK_LONG_DESCRIPTION)
def health_check(args):  # noqa; pylint: disable=unused-argument
    to_remove = []
    data = get_user_config_repo_data()
    if not data:
        return
    for uuid, path in data.get('repos', {}).items():
        if not os.path.isdir(path):
            print('Removing repository {uuid} => {path}'.format(uuid=uuid, path=path))
            to_remove.append(uuid)
    for uuid in to_remove:
        data.get('repos', {}).pop(uuid, None)
    with open(get_user_config_repos(), 'w') as f:
        json.dump(data, f)


_LIST_LONG_DESCRIPTION = '''
Lists repositories on the system. By default it will list all registered user repositories.

It also can list global repositories on the system which usually reside in /usr/share/leapp-repository/
by using the --global commandline flag.

When using the --all commandline flag, all repositories user and globally are listed.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@repo.command('list', help='Lists repositories', description=_LIST_LONG_DESCRIPTION)
@command_opt('global', is_flag=True, help='List globally available repositories only.')
@command_opt('all', is_flag=True, help='List all available user and global repositories.')
def list_repos(args):
    global_repos = {}
    if getattr(args, 'global', None) or args.all:
        global_repos = get_global_repositories_data()

    for entry in global_repos.values():
        print('{name:<35} [{uuid}] => {path}'.format(name=entry['name'], path=entry['path'], uuid=entry['id']))

    if not getattr(args, 'global', None):
        user_repos = get_user_config_repo_data()
        for path in user_repos.get('repos', {}).values():
            if os.path.isdir(path):
                name = get_repository_name(path)
                uuid = get_repository_id(path)
                print('{name:<35} [{uuid}] => {path}'.format(name=name, path=path, uuid=uuid))


def register_path(path):
    """
    Calling this function will register a path to be a well
    :param path: Path to the repository repository
    :return:
    """
    path = os.path.abspath(os.path.realpath(path))
    data = {}
    repos = get_user_config_repos()
    if os.path.isfile(repos):
        with open(repos) as f:
            data = json.load(f)
    data.setdefault('repos', {}).update({get_repository_id(path): path})
    with open(repos, 'w') as f:
        json.dump(data, f)


_REGISTER_LONG_DESCRIPTION = '''
Registers the current user repository in the users repository registry.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@repo.command('register', help='Registers the current repository in the user repository registry.',
              description=_REGISTER_LONG_DESCRIPTION)
@requires_repository
def register_repo(args):  # noqa; pylint: disable=unused-argument
    base_dir = find_repository_basedir('.')
    if base_dir:
        register_path(base_dir)
        print('Repository successfully registered')


_LINK_LONG_DESCRIPTION = '''
Links a given repository to the current repository.

Linking a repository is needed, when the current repository requires things like
Tags, Models, Topics, Workflows etc from another repository and needs to be executable
with `snactor`. Snactor does not know otherwise that it will need to load the content
from another repository. Linking the repositories will make snactor load the items
from the linked repositories.

Repositories can be linked by path, name and repository id.

When using the repository name, beware that the first matching name will be linked.
Therefore it's recommended to rather link repositories by path or repository id.

Usage:
    $ snactor repo link --path ../../other-repository

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@repo.command('link', help='Links a given repository to the current one', description=_LINK_LONG_DESCRIPTION)
@command_opt('path', help='Path to the repository to link')
@command_opt('name', help='Name of the repository to link')
@command_opt('uuid', help='UUID of the repository to link', )
@requires_repository
def link_repo(args):
    if not any((args.path, args.name, args.uuid)):
        raise UsageError('Please specify either --path, --name or --uuid to link another repository.')
    data = get_user_config_repo_data()
    path = args.path
    if not path:
        if args.uuid:
            path = data.get('repos', {}).get(args.uuid, None)
        elif args.name:
            for repository_path in data.get('repos', {}).values():
                if os.path.isdir(repository_path):
                    if args.name == get_repository_name(repository_path):
                        path = repository_path
                        break
    if not path:
        raise UsageError('Please specify a valid repository name, uuid or path')

    if add_repository_link('.', get_repository_id(path)):
        print('Added link to repository {path} - {name}'.format(path=path, name=get_repository_name(path)))


_FIND_LONG_DESCRIPTION = '''
Searches for all repositories and registers them.

When not specifying --path - It will search from the current working directory for existing
leapp repositories and registers all found repositories with the users repository registration.

By using --skip-registration it can be used to just detect repositories without registering them.

If another path should be scanned than the current working directory pass the --path flag.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@repo.command('find', help='Find and registers all repositories')
@command_opt('skip-registration', is_flag=True, help='Do not register discovered repositories.')
@command_opt('path', help='Path to scan from - If not specified the current working directory is assumed')
def find_repositories(args):
    path = args.path or os.path.realpath('.')
    result = subprocess.check_output(['/usr/bin/find', '-L', path, '-name', '.leapp']).decode('utf-8')
    for repository in result.split('\n'):
        if repository.strip():
            repository = os.path.dirname(repository)
            if not args.skip_registration:
                print('Registering {path}'.format(path=repository))
                register_path(repository)
            else:
                print(repository)


_REPOSITORY_CONFIG = '''
[repositories]
repo_path=${repository:root_dir}

[database]
path=${repository:state_dir}/leapp.db
'''
_LONG_DESCRIPTION = '''
Creates a new local repository for writing Actors, Models, Tags,
Topics, and Workflows or adding shared files, tools or libraries.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@repo.command('new', help='Creates a new repository', description=_LONG_DESCRIPTION)
@command_arg('name')
def new_repository(args):
    name = args.name
    basedir = os.path.join('.', name)
    if os.path.isdir(basedir):
        raise CommandError("Directory already exists: {}".format(basedir))

    os.mkdir(basedir)
    repository_dir = os.path.join(basedir, '.leapp')
    os.mkdir(repository_dir)
    with open(os.path.join(repository_dir, 'info'), 'w') as f:
        json.dump({
            'name': name
        }, f)
    with open(os.path.join(repository_dir, 'leapp.conf'), 'w') as f:
        f.write(_REPOSITORY_CONFIG)

    register_path(basedir)
    sys.stdout.write("New repository {} has been created in {}\n".format(name, os.path.realpath(name)))



from __future__ import print_function
from multiprocessing import util, Process, Manager

import os
import pytest

import leapp  # noqa: F401; pylint: disable=unused-import
from leapp.utils.workarounds import fqdn


def test_mp_is_patched():

    def child_fun(_lst):
        pid = os.fork()
        if not pid:
            os.execvpe('whatever', [], os.environ)
        else:
            os.wait()

    m = Manager()
    lst = m.list()

    p = Process(target=child_fun, args=(lst,))
    p.start()
    p.join()
    if lst:
        for el in lst:
            print(el)


def test_mp_workaround_applied():
    if getattr(util, 'os', None) is None:
        assert util.Finalize.__name__ == 'FixedFinalize'


@pytest.mark.parametrize(
    ('input_fqdn', 'valid'),
    [
        ('foo.bar.com', True),
        ('foo\xa0bar.foo.com', False),
        ('-foo.bar', False),
        ('foo.bar-', False),
        ('foo.-bar.1234', False),
        ('a1.b2.c3', True),
        ('123.foo.bar', True),
        ('123.f-o-o.b-a-r', True),
    ]
)
def test_fqdn_is_patched(input_fqdn, valid):

    def getfqdn():
        return input_fqdn

    fn = fqdn.decorate_getfqdn(getfqdn)
    expected_fqdn = input_fqdn if valid else 'invalid.hostname'
    assert fn() == expected_fqdn



from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.tags import FirstPhaseTag, WorkflowApiTestWorkflowTag
from leapp.workflows.api.v3 import TestAPI


class ApiV3Test(Actor):
    """
    This actor will check that the testing API ApiV3Test is behaving as expected.
    """

    name = 'api_v3_test'
    consumes = ()
    produces = ()
    apis = (TestAPI,)
    tags = (FirstPhaseTag, WorkflowApiTestWorkflowTag)

    def process(self):
        api = TestAPI()
        # Expected output is always the order a, b, c
        # The parameter order here is however b, c, a
        if api.order_change(2, 3, 1) != (1, 2, 3):
            raise StopActorExecutionError("order change API failure")

        # Expected is that the input is prefixed with 'survivor.v3.'
        if api.survivor("APIV3Test") != "survivor.v3.APIV3Test":
            raise StopActorExecutionError("v3 survivor test failure")



import os
import socket
import sys
import types
import uuid
from multiprocessing import Process, Queue

import pytest
from _pytest.python import pytest_pyfunc_call as original_pytest_pyfunc_call

from leapp.compat import raise_with_traceback
from leapp.messaging.inprocess import InProcessMessaging
from leapp.repository.scan import find_and_scan_repositories
from leapp.utils import get_api_models
from leapp.utils.audit import Execution, get_connection
from leapp.utils.repository import find_repository_basedir


PY27 = sys.version_info[:2] == (2, 7)
PY36 = sys.version_info[:2] == (3, 6)
PY38 = sys.version_info[:2] == (3, 8)
HIGHER = sys.version_info[:2] > (3, 8)


def _patched_name_py27(code, name):
    return types.CodeType(
        code.co_argcount,
        code.co_nlocals,
        code.co_stacksize,
        code.co_flags,
        code.co_code,
        code.co_consts,
        code.co_names,
        code.co_varnames,
        code.co_filename,
        name,
        code.co_firstlineno,
        code.co_lnotab,
        code.co_freevars,
        code.co_cellvars,
    )


def _patched_name_py36(code, name):
    return types.CodeType(
        code.co_argcount,
        code.co_kwonlyargcount,
        code.co_nlocals,
        code.co_stacksize,
        code.co_flags,
        code.co_code,
        code.co_consts,
        code.co_names,
        code.co_varnames,
        code.co_filename,
        name,
        code.co_firstlineno,
        code.co_lnotab,
        code.co_freevars,
        code.co_cellvars,

    )


def _patched_name_py38(code, name):
    code.replace(co_name=name)
    return code


def _tb_pack(tb):
    result = []
    while tb is not None:
        code = tb.tb_frame.f_code
        result.append((code.co_filename, tb.tb_lineno, code.co_name))
        tb = tb.tb_next
    return result


def _tb_unpack(packed):
    previous = '_e'
    globs = {previous: StopIteration}
    for i, (filename, lineno, name) in enumerate(reversed(packed)):
        current = '_%d' % i
        eval(compile(  # noqa; pylint: disable=eval-used
            '%sdef %s(): raise %s()' % ('\n' * (lineno - 1), current, previous), filename, 'exec'), globs)
        previous = current
        func = globs[current]
        if PY27:
            func.func_code = _patched_name_py27(func.func_code, name)
        elif PY36:
            func.__code__ = _patched_name_py36(func.__code__, name)
        elif PY38 or HIGHER:
            func.__code__ = _patched_name_py38(func.__code__, name)
        else:
            raise NotImplementedError(
                "Traceback forwarding is not implemented for your"
                " Python version. Currently supported 2.7, 3.6, 3.8"
                " and higher"
            )
    try:
        globs[previous]()
    except StopIteration:
        globs.clear()
        return sys.exc_info()[2].tb_next


class ActorContext(object):
    """
    ActorContext is a helper class for testing actors. It helps to eliminate the boilerplate for executing
    actors. It provides a set of methods that allow specifying input messages for the actor, executing the
    actor and to retrieve messages sent by the actor.
    """
    apis = ()

    def __init__(self, actor=None):
        self._actor = actor
        self._messaging = InProcessMessaging()

    def set_actor(self, actor):
        """
        Internally used method to set the current actor specification object to setup the current actor for the
        test function.

        :param actor: ActorSpecification instance to use.
        :return: None
        """
        type(self).name = actor.name + '_feeder'
        # Consumes is here what it actually produces because we want to consume produced messages in 'consume'
        type(self).consumes = get_api_models(actor, 'produces')
        self._actor = actor

    def feed(self, *models):
        """
        Feed the messaging model with messages to be available to consume.

        :param models: Data in form of model instances to be available for the actor to consume.
        :type models: Variable number of instances of classes derived from :py:class:`leapp.models.Model`

        :return: None
        """
        for model in models:
            self._messaging.feed(model, self)

    def run(self, config_model=None):
        """
        Execute the current actor.

        :param config_model: Config model for the actor to consume.
        :type config_model: Config model instance derived from :py:class:`leapp.models.Model`
        :return: None
        """
        config_model_cls = config_model.__class__ if config_model else None
        if config_model:
            self._messaging.feed(config_model, self)
            # we have to make messaging system aware of config model being used as this is normally done by workflow
            self._messaging._config_models = (config_model_cls,)
        # the same as above applies here for actor
        self._actor(messaging=self._messaging, config_model=config_model_cls).run()

    def messages(self):
        """
        Returns raw messages produced by the actor.

        :return: list of raw message data dictionaries.
        """
        return self._messaging.messages()

    def consume(self, *models):
        """
        Retrieve messages produced by the actor execution and specified in the actors :py:attr:`produces`
        attribute, and filter message types by models.

        :param models: Models to use as a filter for the messages to return
        :type models: Variable number of the derived classes from :py:class:`leapp.models.Model`
        :return:
        """
        return tuple(self._messaging.consume(self, *models))


@pytest.fixture(scope='module')
def loaded_leapp_repository(request):
    """
    This fixture will ensure that the repository for the current test run is loaded with all its links etc.

    This enables running actors and using models, tags, topics, workflows etc.

    Additionally loaded_leapp_repository gives you access to a :py:class:`leapp.repository.manager.RepositoryManager`
    instance.

    :Example:

    .. code-block:: python

        from leapp.snactor.fixture import loaded_leapp_repository
        from leapp.models import ExampleModel, ProcessedExampleModel

        def my_repository_library_test(loaded_leapp_repository):
            from leapp.libraries.common import global
            e = ExampleModel(value='Some string')
            result = global.process_function(e)
            assert type(result) is ProcessedExampleModel

    """
    repository_path = find_repository_basedir(request.module.__file__)
    os.environ['LEAPP_CONFIG'] = os.path.join(repository_path, '.leapp', 'leapp.conf')
    os.environ['LEAPP_HOSTNAME'] = socket.getfqdn()
    context = str(uuid.uuid4())
    with get_connection(None):
        Execution(context=context, kind='snactor-test-run', configuration='').store()
        os.environ["LEAPP_EXECUTION_ID"] = context

        manager = getattr(request.session, 'leapp_repository', None)
        if not manager:
            manager = find_and_scan_repositories(repository_path, include_locals=True)
            manager.load(resolve=True)
        yield manager


@pytest.fixture(scope='function')
def current_actor_context(loaded_leapp_repository):
    """
    This fixture will prepare an environment for the actor the test belongs to, to be safely executable.

    current_actor_context Is an instance of :py:class:`leapp.snactor.fixture.ActorContext` and gives access
    to its methods for feeding an actor with input data, running the actor, and retrieving messages produced
    by the actor during its execution.


    :Example:

    .. code-block:: python

        from leapp.snactor.fixture import current_actor_context
        from leapp.models import ConsumedExampleModel, ProducedExampleModel

        def test_actor_lib_some_function(current_actor_context):
            # Feed with messages to be consumable by the actor that is going to be executed.
            current_actor_context.feed(ConsumedExampleModel(value='Some random data'))

            # Execute the actor
            current_actor_context.run()

            # Ensure that at least one message is produced
            assert current_actor_context.consume(ProducedExampleModel)

            # Ensure the value is what we expect
            assert current_actor_context.consume(ProducedExampleModel)[0].value == 42
    """
    return type('CurrentActorContext', (ActorContext,), {'repository': loaded_leapp_repository, 'name': None})()


@pytest.fixture(scope='function')
def current_actor_libraries(request, loaded_leapp_repository):
    """
    This fixture will make libraries that are private to the actor only available only for the scope of the
    test function that uses this fixture.

    :Example:

    .. code-block:: python

        from leapp.snactor.fixture import current_actor_libraries

        def test_actor_lib_some_function(current_actor_libraries):
            from leapp.libraries.actor import private
            assert private.some_function(1) == 42

    """
    actor = _get_actor(request.module, loaded_leapp_repository)
    with actor.injected_context():
        yield actor


def _get_actor(module, repository):
    """
    Looks up the current actor based on the module passed. With help of the location of the module where
    the tests reside, the actor that is requested can be deduced since the full path of the actor should be
    the prefix of the path for the current test module.

    :param module: A python module object in which reside the tests for the actor to be run.
    :param repository: Instance of a :py:class:`leapp.repository.manager.RepositoryManager`
    :return: ActorDefinition instance or None, when the actor could not be found
    """
    path = os.path.realpath(module.__file__)
    for actor in repository.actors:
        if path.startswith(os.path.realpath(actor.full_path) + os.sep):
            return actor
    return None


@pytest.fixture(scope='module')
def leapp_forked():
    pass


def _execute_test(q, pyfuncitem):
    """
    This function is called in the child process from pytest_pyfunc_call via multiprocessing.Process.

    :param q: A multiprocessing.Queue object to pass data to the parent process
    :param pyfuncitem: pytest item describing the current test function
    :return: None
    """
    try:
        if 'current_actor_context' in pyfuncitem.funcargs:
            actor = _get_actor(pyfuncitem.module, pyfuncitem.funcargs['current_actor_context'].repository)
            pyfuncitem.funcargs['current_actor_context'].set_actor(actor)
        original_pytest_pyfunc_call(pyfuncitem=pyfuncitem)
        q.put((True, None))
    except BaseException:  # noqa; pylint: disable=broad-except
        # We need this broad exception to catch all errors and pass them through to the parent process
        e_type, e_exc, e_tb = sys.exc_info()
        q.put((False, (e_type, e_exc, _tb_pack(e_tb))))


if hasattr(pytest, 'hookimpl'):
    @pytest.hookimpl(tryfirst=True)
    def pytest_pyfunc_call(pyfuncitem):
        """
        This function is a hook for pytest implementing the ability to run the actors in tests safely.

        It will call :py:func:`leapp.snactor.fixture._execute_test` in a child process if the current test uses the
        :py:func:`current_actor_context` fixture. If it doesn't use the :py:func:`current_actor_context` fixture, it
        will default to the default `pytest_pyfunc_call` implementation.
        """
        if not any(arg in pyfuncitem.funcargs for arg in ('current_actor_context', 'leapp_forked')):
            return None
        q = Queue()
        p = Process(target=_execute_test, args=(q, pyfuncitem))
        p.start()
        p.join()

        # Ensure we are actually getting a result - Otherwise ensure this is marked as failure
        assert not q.empty()
        r, e = q.get()
        if e:
            raise_with_traceback(e[1], _tb_unpack(e[2]))
        return r



class WorkflowCommand(object):
    def __init__(self, command, arguments):
        self._command = command
        self._arguments = arguments

    def encode(self):
        return {'command': self._command, 'arguments': self._arguments}


class SkipPhasesUntilCommand(WorkflowCommand):
    COMMAND = 'skip_phases_until'

    def __init__(self, until_phase):
        super(SkipPhasesUntilCommand, self).__init__(self.COMMAND, {'until_phase': until_phase})



import pytest
import six

from leapp.libraries.stdlib import run


def test_invalid_command():
    """
    Test raising TypeError when the command is not a tuple or list.

    :return: Pass/Fail
    """
    with pytest.raises(TypeError):
        run('i should be a list or tuple really')


def test_split():
    """
    Test the output lines are split into lists.

    :return: Pass/Fail
    """
    cmd = ['ls']
    result = run(cmd, split=True)
    assert isinstance(result['stdout'], list)
    assert len(result['stdout']) > 1


def test_no_encoding():
    """
    Test the output is base64 encoded after getting binary data

    :return: Pass/Fail
    """
    cmd = ['echo', '-n', '-e', '\\xeb']
    result = run(cmd, encoding=None)
    assert isinstance(result['stdout'], six.binary_type)



from collections import namedtuple
import datetime
import json
import os
import tempfile

import jsonschema
import pytest

from leapp.reporting import (
    _add_to_dict,
    _DEPRECATION_SEVERITY_THRESHOLD,
    create_report_from_deprecation,
    create_report_from_error,
    _create_report_object,
    Audience, Flags, Groups, Key, RelatedResource, Summary, Severity, Tags, Title
)
from leapp.utils.report import generate_report_file


ReportPrimitive = namedtuple('ReportPrimitive', ['data', 'path', 'value', 'is_leaf_list'])

leaf_is_dict = ReportPrimitive({}, ('summary',), 'We better fix this!', False)
leaf_is_list = ReportPrimitive({}, ('summary',), 'We better fix this!', False)

path_taken = ReportPrimitive({'summary': 'Fix is coming!'}, ('summary',), 'We better fix this!', False)

leaf_list_append1 = ReportPrimitive(
    {}, ('detail', 'remediation'), {'type': 'hint', 'context': 'Read the docs!'}, True
)
leaf_list_append2 = ReportPrimitive(
    {}, ('detail', 'remediation'), {'type': 'command', 'context': 'man leapp'}, True
)


def value_from_path(data, path):
    value = data
    for p in path:
        if isinstance(value, dict) and value.get(p, None):
            value = value[p]
        else:
            return None
    return value


def is_path_valid(data, path):
    return bool(value_from_path(data, path))


def assert_leaf_list(data, path, is_leaf_list):
    res = isinstance(value_from_path(data, path), list)
    if is_leaf_list:
        assert res
    else:
        assert not res


@pytest.mark.parametrize('primitive', (leaf_is_dict, leaf_is_list))
def test_add_to_dict_func_leaves(primitive):
    data, path, value, is_leaf_list = primitive

    _add_to_dict(data, path, value, is_leaf_list)
    assert is_path_valid(data, path)
    assert_leaf_list(data, path, is_leaf_list)


def test_add_to_dict_func_path_taken():
    data, path, value, is_leaf_list = path_taken

    with pytest.raises(ValueError):
        _add_to_dict(data, path, value, is_leaf_list)


def test_add_to_dict_func_append():
    data, path, value, is_leaf_list = leaf_list_append1

    _add_to_dict(data, path, value, is_leaf_list)
    assert is_path_valid(data, path)
    assert_leaf_list(data, path, is_leaf_list)

    # we do not want data again as we will re-use the dict for appending
    _data, path, value, is_leaf_list = leaf_list_append2
    _add_to_dict(data, path, value, is_leaf_list)

    assert is_path_valid(data, path)
    assert_leaf_list(data, path, is_leaf_list)
    assert len(value_from_path(data, path)) == 2


def test_report_groups():
    assert Groups.INHIBITOR == "inhibitor"
    assert Groups.REPOSITORY == "repository"
    # make sure you can use a custom group in pressing need
    assert Groups(["This is a new group", Groups.INHIBITOR]).value == ["This is a new group", "inhibitor"]


def test_report_tags_and_flags():
    # For compatibility purposes make sure that those primitives are still available to be used by actor writers
    assert Flags.INHIBITOR == "inhibitor"
    assert Tags.REPOSITORY == "repository"
    # make sure you can use a custom flag in pressing need
    assert Flags(["This is a new flag", Groups.INHIBITOR]).value == ["This is a new flag", "inhibitor"]


def test_convert_from_error_to_report():
    error_dict_no_details = {
        'message': 'The system is not registered or subscribed.',
        'time': '2019-11-19T05:13:04.447562Z',
        'details': None,
        'actor': 'verify_check_results',
        'severity': 'error'}
    report = create_report_from_error(error_dict_no_details)
    assert report['severity'] == 'high'
    assert report['title'] == 'The system is not registered or subscribed.'
    assert report['audience'] == 'sysadmin'
    assert report['summary'] == ''
    error_dict_with_details = {
        'message': 'The system is not registered or subscribed.',
        'time': '2019-11-19T05:13:04.447562Z',
        'details': 'Some other info that should go to report summary',
        'actor': 'verify_check_results',
        'severity': 'error'}
    report = create_report_from_error(error_dict_with_details)
    assert report['severity'] == 'high'
    assert report['title'] == 'The system is not registered or subscribed.'
    assert report['audience'] == 'sysadmin'
    assert report['summary'] == 'Some other info that should go to report summary'
    # make sure key argument is there
    error_dict_key_not_specified = {
        'message': 'The system is not registered or subscribed.',
        'time': '2019-11-19T05:13:04.447562Z',
        'details': 'Some other info that should go to report summary',
        'actor': 'verify_check_results',
        'severity': 'error'}
    report = create_report_from_error(error_dict_key_not_specified)
    assert "key" in report


def test_create_report_from_deprecation():
    curr_date = datetime.datetime.today()
    old_date = curr_date - _DEPRECATION_SEVERITY_THRESHOLD - datetime.timedelta(days=1)

    for (since, severity) in ((curr_date, 'medium'), (old_date, 'high')):
        data = {
            'message': 'Usage of deprecated Model "DeprecatedModel"',
            'filename': '/etc/leapp/repos.d/system_upgrade/el7toel8/actors/fooactor/actor.py',
            'line': '        self.produce(DeprecatedModel(foo=bar))\n',
            'lineno': 51,
            'since': since.strftime('%Y-%m-%d'),
            'reason': 'The DeprecatedModel has been deprecated.'}
        report = create_report_from_deprecation(data)
        # I do not want to just copy-paste the code from the function, so let's
        # do the check in this way
        assert data['message'] in report['title']
        assert '{}:{}'.format(data['filename'], data['lineno']) in report['title']
        assert '{}:{}'.format(data['filename'], data['lineno']) in report['summary']
        assert report['severity'] == severity
        assert report['audience'] == 'developer'
        assert data['reason'] in report['summary']
        assert data['since'] in report['summary']
        assert data['line'] in report['summary']
        # make sure key argument is there
        assert 'key' in report


def test_create_report_stable_key(monkeypatch):
    report_entries1 = [Title('Some report title'), Summary('Some summary not used for dynamical key generation'),
                       Audience('sysadmin')]
    report_entries2 = [Title('Some report title'),
                       Summary('Another summary not used for dynamical key generation'), Audience('sysadmin')]
    report_entries_with_severity = [Title('Some report title'),
                                    Summary('Another summary not used for dynamical key generation'),
                                    Severity('high')]
    report_entries_with_audience = [Title('Some report title'),
                                    Summary('Another summary not used for dynamical key generation'),
                                    Audience('developer')]
    report_entries_fixed_key = [Title('Some report title'), Summary('Different summary'), Key('42')]
    report_entries_dialog1 = [Title('Some report title'), RelatedResource('dialog', 'unanswered_section'),
                              Audience('sysadmin'), Summary('Summary')]
    report_entries_dialog_key_set = [Title('Some report title'), RelatedResource('dialog', 'unanswered_section'),
                                     RelatedResource('dialog', 'another_unanswered_section'), Audience('sysadmin'),
                                     Summary('Summary'), Key('dialogkey42')]
    report1 = _create_report_object(report_entries1)
    report2 = _create_report_object(report_entries2)
    report_with_severity = _create_report_object(report_entries_with_severity)
    report_with_audience = _create_report_object(report_entries_with_audience)
    report_fixed_key = _create_report_object(report_entries_fixed_key)
    report_dialog1 = _create_report_object(report_entries_dialog1)
    report_dialog_key_set = _create_report_object(report_entries_dialog_key_set)
    # check that all reports have key field
    for report in [report1, report2, report_with_severity, report_with_audience, report_fixed_key]:
        assert "key" in report.report  # pylint: disable = unsupported-membership-test
    # check that reports with same title but different summary have same dynamically generated key
    assert report1.report["key"] == report2.report["key"]
    # check that entries different in severity only will have different generated keys
    assert report2.report["severity"] != report_with_severity.report["severity"]
    assert report2.report["key"] != report_with_severity.report["key"]
    # check that entries different in audience only will have different generated keys
    assert report2.report["audience"] != report_with_audience.report["audience"]
    assert report2.report["key"] != report_with_audience.report["key"]
    # check that key from provided Key entry is taken
    assert report_fixed_key.report["key"] == "42"
    # check that specific dialog related resources are not considered in key generation
    assert report_dialog1.report["key"] == report1.report["key"]
    # check that providing a key for the dialog report is a solution
    assert report_dialog1.report["key"] != report_dialog_key_set.report["key"]
    assert report_dialog_key_set.report["key"] == "dialogkey42"
    # check that empty string can't serve as report Key
    with pytest.raises(ValueError) as err:
        _create_report_object([Title('A title'), Summary('A summary'), Key('')])
    assert str(err.value) == 'Key can not be an empty string.'
    # check that if LEAPP_DEVEL_FIXED_REPORT_KEY is set and there is no key in the report - a ValueError is raised
    monkeypatch.setenv('LEAPP_DEVEL_FIXED_REPORT_KEY', '1')
    with pytest.raises(ValueError):
        _create_report_object([Title('A title'), Summary('A summary')])
    # check that Key accepts string parameters only
    for bad_uuid in [42, object(), 42.42]:
        with pytest.raises(ValueError) as err:
            Key(bad_uuid)
        assert str(err.value) == 'Key value should be a string.'


def test_report_jsonschema():
    reports_dir = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        'data',
        'report-schema',
    )

    def _load_json(a_file):
        with open(a_file) as f:
            return json.loads(f.read())

    for report in [r for r in os.listdir(reports_dir) if os.path.isfile(r)]:
        report_data = _load_json(os.path.join(reports_dir, report))
        for schema in ['report-schema-v100.json', 'report-schema-v110.json']:
            # schemas directory contain a symlink to report json-schemas
            schema_data = _load_json(os.path.join(reports_dir, 'schemas', schema))
            jsonschema.validate(report_data, schema_data)


def test_report_backwards_compatibility():
    msg1 = {u'groups': [u'selinux', u'security'],
            u'title': u'SElinux relabeling has been scheduled',
            'timeStamp': u'2021-02-10T13:13:40.576437Z',
            'hostname': u'ivasilev-tagsflagsgroups',
            'actor': u'check_se_linux',
            u'summary': u'SElinux relabeling has been scheduled as the status was permissive/enforcing.',
            u'audience': u'sysadmin',
            u'key': u'c12a05a22be0b5bc0af3f1119898ea6d8639d9c4',
            'id': '29279fc2e3100cfdf8ac30f48b293185192672d5a5ae9e0cfbc76cbfc919d807',
            u'severity': u'info'}
    report = [msg1]
    # make sure normal mode works
    for report_format in ['.json', '.txt']:
        with tempfile.NamedTemporaryFile(suffix=report_format) as reportfile:
            generate_report_file(report, 'leapp-run-id', reportfile.name, '1.1.0')
    # make sure report output conversion to specific version works as well
    for report_format in ['.json', '.txt']:
        with tempfile.NamedTemporaryFile(suffix=report_format) as reportfile:
            generate_report_file(report, 'leapp-run-id', reportfile.name, '1.0.0')
            with open(reportfile.name) as f:
                data = f.read()
            if report_format == '.json':
                a_report = json.loads(data)
                msg = a_report['entries'][0]
                assert 'key' not in msg
            else:
                assert 'Key' not in data



from __future__ import print_function

import codecs
import os

from leapp.compat import string_types
from leapp.libraries.stdlib.eventloop import POLL_HUP, POLL_IN, POLL_OUT, POLL_PRI, EventLoop

STDIN = 0
STDOUT = 1
STDERR = 2


def _multiplex(ep, read_fds, callback_raw, callback_linebuffered,
               encoding='utf-8', write=None, timeout=1, buffer_size=80):
    # Register the file descriptors (stdout + stderr) with the epoll object
    # so that we'll get notifications when data are ready to read
    for fd in read_fds:
        ep.register(fd, POLL_IN | POLL_PRI)

    # Register a write file descriptor
    if write:
        ep.register(write[0], POLL_OUT)

    # Offset into the `write[1]` buffer where we should continue writing to stdin
    offset = 0

    # We need to keep track of which file descriptors have already been drained
    # because when running under `pytest` it seems that all `epoll` events are
    # received twice so using solely `ep.unregister(fd)` will not work
    hupped = set()
    # Total number of 'hupped' file descriptors we expect
    num_expected = len(read_fds) + (1 if write else 0)
    # Set up file-descriptor specific buffers where we'll buffer the output
    buf = {fd: bytes() for fd in read_fds}
    if encoding:
        linebufs = {fd: '' for fd in read_fds}
        decoders = {fd: codecs.getincrementaldecoder(encoding)() for fd in read_fds}

    def _get_fd_type(fd):
        """
        File descriptors passed via `read_fds` are always representing [stdout, stderr],
        since arrays start at index 0, we need to add 1 to get the real symbolic value
        `STDOUT` or `STDERR`.
        """
        return read_fds.index(fd) + 1

    while not ep.closed and len(hupped) != num_expected:
        events = ep.poll(timeout)
        for fd, event in events:
            if event == POLL_HUP:
                hupped.add(fd)
                ep.unregister(fd)
            if event & (POLL_IN | POLL_PRI) != 0:
                fd_type = _get_fd_type(fd)
                read = os.read(fd, buffer_size)
                callback_raw((fd, fd_type), read)
                if encoding:
                    linebufs[fd] += decoders[fd].decode(read)
                    while '\n' in linebufs[fd]:
                        pre, post = linebufs[fd].split('\n', 1)
                        linebufs[fd] = post
                        callback_linebuffered((fd, fd_type), pre)
                buf[fd] += read
            elif event == POLL_OUT:
                # Write data to pipe, `os.write` returns the number of bytes written,
                # thus we need to offset
                wfd, data = write
                if fd in hupped:
                    continue
                offset += os.write(fd, data[offset:])
                if offset == len(data):
                    hupped.add(fd)
                    ep.unregister(fd)
                    os.close(fd)

    # Process leftovers from line buffering
    if encoding:
        for (fd, lb) in linebufs.items():
            if lb:
                # [stdout, stderr] is relayed, stdout=1 a stderr=2
                # as the field starting indexed is 0, so the +1 needs to be added
                callback_linebuffered((fd, _get_fd_type(fd)), lb)

    return buf


def _call(command, callback_raw=lambda fd, value: None, callback_linebuffered=lambda fd, value: None,
          encoding='utf-8', poll_timeout=1, read_buffer_size=80, stdin=None, env=None):
    """
        :param command: The command to execute
        :type command: list, tuple
        :param encoding: Decode output or encode input using this encoding
        :type encoding: str
        :param poll_timeout: Timeout used by epoll to wait certain amount of time for activity on file descriptors
        :type poll_timeout: int
        :param read_buffer_size: How much data are we going to read from the file descriptors each iteration.
                                 The default value of 80 chosen to correspond with suggested terminal line width
        :type read_buffer_size: int
        :param callback_raw: Callback executed on raw data (before decoding) as they are read from file descriptors
        :type callback_raw: ((fd: int, fd_type: int), buffer: bytes) -> None
        :param callback_linebuffered: Callback executed on decoded lines as they are read from the file descriptors
        :type callback_linebuffered: ((fd: int, fd_type: int), value: str) -> None
        :param stdin: String or a file descriptor that will be written to stdin of the child process
        :type stdin: int, str
        :param env: Environment variables to use for execution of the command
        :type env: dict
        :return: {'stdout' : stdout, 'stderr': stderr, 'signal': signal, 'exit_code': exit_code, 'pid': pid}
        :rtype: dict
    """
    if not isinstance(command, (list, tuple)):
        raise TypeError('command parameter has to be a list or tuple')
    if not callable(callback_raw) or\
            (getattr(callback_raw, '__code__', None) and callback_raw.__code__.co_argcount != 2):
        raise TypeError('callback_raw parameter has to be callable accepting 2 parameters')
    if (not callable(callback_linebuffered) or (getattr(callback_linebuffered, '__code__', None) and  # noqa
                                                callback_linebuffered.__code__.co_argcount != 2)):
        raise TypeError('callback_linebuffered parameter has to be callable accepting 2 parameters')
    if not isinstance(poll_timeout, int) or isinstance(poll_timeout, bool) or poll_timeout <= 0:
        raise ValueError('poll_timeout parameter has to be integer greater than zero')
    if not isinstance(read_buffer_size, int) or isinstance(read_buffer_size, bool) or read_buffer_size <= 0:
        raise ValueError('read_buffer_size parameter has to be integer greater than zero')

    environ = os.environ.copy()
    if env:
        if not isinstance(env, dict):
            raise TypeError('env parameter has to be a dictionary')
        environ.update(env)
    # Create a separate pipe for stdout/stderr
    #
    # The parent process is going to use the read-end of the pipes for reading child's
    # stdout/stderr, whereas the forked children process is going to use the write-end
    # of the pipes to pass data to parent
    stdout, wstdout = os.pipe()
    stderr, wstderr = os.pipe()

    # We allow stdin to be either a file descriptor (int) or a string and we need to handle
    # each of those cases differently
    #
    # The case where stdin is a file descriptor is simple -- we just need to dup2() the file
    # descriptor into the child process' stdin. If stdin is a string, though, the situation is
    # more complicated and we need to create another pipe and write the string to the pipe
    # in the _multiplex function
    fstdin, wstdin = None, None
    stdin_fd, stdin_str = False, False
    if isinstance(stdin, int):
        stdin_fd = True
    elif isinstance(stdin, string_types):
        stdin_str = True
        fstdin, wstdin = os.pipe()
    elif stdin is not None:
        raise TypeError('stdin has to be either a file descriptor or string, not "{!s}"'.format(type(stdin)))

    ep = EventLoop()
    pid = os.fork()
    if pid > 0:
        # Since pid > 0, we are in the parent process, so we have to close the write-end
        # file descriptors
        os.close(wstdout)
        os.close(wstderr)
        # Extra optional arguments for the `_multiplex` function
        extra = {}
        if stdin_str:
            # NOTE: We use the same encoding for encoding the stdin string as well which might
            # be suboptimal in certain cases -- there are two possible solutions:
            #  1) Rather than string require the `stdin` parameter to already be bytes()
            #  2) Add another parameter for stdin_encoding
            extra['write'] = (wstdin, stdin.encode(encoding))
            os.close(fstdin)

        read = _multiplex(
            ep,
            [stdout, stderr],
            callback_raw,
            callback_linebuffered,
            timeout=poll_timeout,
            buffer_size=read_buffer_size,
            encoding=encoding,
            **extra
        )

        # Wait for the child to finish
        pid, status = os.wait()
        ep.close()

        # The status variable is a 16 bit value, where the lower octet describes
        # the signal which killed the process, and the upper octet is the exit code
        signal, exit_code = status & 0xff, status >> 8 & 0xff
        ret = {'signal': signal, 'exit_code': exit_code, 'pid': pid}
        if not encoding:
            ret.update({
                'stdout': read[stdout],
                'stderr': read[stderr]
            })
        else:
            ret.update({
                'stdout': read[stdout].decode(encoding),
                'stderr': read[stderr].decode(encoding)
            })
        return ret
    if pid == 0:
        # We are in the child process, so we need to close the read-end of the pipes
        # and assign our pipe's file descriptors to stdout/stderr
        #
        # If `stdin` is specified as a file descriptor, we simply pass it as the stdin of the
        # child. In case `stdin` is specified as a string, we pass in the read end of our
        # stdin pipe
        if stdin_fd:
            os.dup2(stdin, STDIN)
        if stdin_str:
            os.close(wstdin)
            os.dup2(fstdin, STDIN)
        os.close(stdout)
        os.close(stderr)
        os.dup2(wstdout, STDOUT)
        os.dup2(wstderr, STDERR)
        os.execvpe(command[0], command, env=environ)



import os
import subprocess
import sys

import pytest
from six import PY3


def _evaluate_results(results, *expected):
    if PY3:
        results = results.decode('utf-8')
    expected = set(expected)
    lines = set(line.split()[-1] for line in results.split('\n') if line.startswith('<<<TEST>>>: '))
    assert lines == expected


_LEAPP_RERUN_CONFIG = None
_LEAPP_DB_PATH = None


def setup_module():
    os.environ['PYTHONDONTWRITEBYTECODE'] = '1'


@pytest.fixture(autouse=True, scope='module')
def create_rerun_tempdir(tmpdir_factory):
    global _LEAPP_RERUN_CONFIG
    global _LEAPP_DB_PATH
    tmpdir = tmpdir_factory.mktemp('leapp-rerun')
    _LEAPP_DB_PATH = str(tmpdir.join('leapp.db'))
    config = tmpdir.join('leappconfig')
    repo_test_path = os.path.join(os.path.dirname(os.path.dirname(
        os.path.realpath(__file__))), 'data/leapp-rerun-tests-repos')
    _LEAPP_RERUN_CONFIG = str(config)
    config.write('''
[repositories]
repo_path={repo_test_path}

[database]
path={test_data_path}/leapp.db

[archive]
dir={test_data_path}/logs/archive

[files_to_archive]
dir={test_data_path}/logs/archive
files=leapp-upgrade.log,leapp-report.json,leapp-report.txt

[logs]
dir={test_data_path}/logs
files=leapp-upgrade.log,leapp-preupgrade.log

[report]
dir={test_data_path}/logs
files=leapp-report.json,leapp-report.txt
answerfile={test_data_path}/logs/answerfile
userchoices={test_data_path}/logs/answerfile.userchoices
    '''.format(test_data_path=str(tmpdir), repo_test_path=repo_test_path))


def _perform_rerun(tags=(), unsupported=True, run_upgrade=True):

    leapp_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
                              'data/leapp-rerun-tests-repos/noroot_leapp.py')
    env = {'LEAPP_CONFIG': _LEAPP_RERUN_CONFIG}
    if os.path.exists(_LEAPP_DB_PATH):
        os.unlink(_LEAPP_DB_PATH)
    args = []

    for tag in tags:
        args.extend(['--only-actors-with-tag', tag])

    try:
        os.environ.update(env)
        if run_upgrade:
            subprocess.check_call([sys.executable, leapp_path, 'upgrade'], env=os.environ)
        if unsupported:
            env['LEAPP_UNSUPPORTED'] = '1'
        os.environ.update(env)
        return subprocess.check_output([sys.executable, leapp_path, 'rerun'] + args + ['FirstBoot'], env=os.environ)
    finally:
        os.environ.pop('LEAPP_CONFIG', None)
        os.environ.pop('LEAPP_UNSUPPORTED', None)


@pytest.mark.skip(reason='Commands have been moved to leapp-repository - The tests will be moved in a separate PR')
def test_leapp_rerun_no_unsupported_environment_var():
    with pytest.raises(subprocess.CalledProcessError):
        _perform_rerun(unsupported=False, run_upgrade=False)


@pytest.mark.skip(reason='Commands have been moved to leapp-repository - The tests will be moved in a separate PR')
def test_leapp_rerun_no_upgrade():
    with pytest.raises(subprocess.CalledProcessError):
        _perform_rerun(run_upgrade=False)


@pytest.mark.skip(reason='Commands have been moved to leapp-repository - The tests will be moved in a separate PR')
def test_leapp_rerun_no_parameters():
    result = _perform_rerun(run_upgrade=True)
    _evaluate_results(result, 'FirstBootActor', 'ReRunActor', 'ReRunActorOther')


@pytest.mark.skip(reason='Commands have been moved to leapp-repository - The tests will be moved in a separate PR')
def test_leapp_rerun_only_actors_with_tag():
    result = _perform_rerun(tags=('ReRunVerifyTag',), run_upgrade=True)
    _evaluate_results(result, 'ReRunActor')


@pytest.mark.skip(reason='Commands have been moved to leapp-repository - The tests will be moved in a separate PR')
def test_leapp_rerun_only_actors_with_multiple_tags():
    result = _perform_rerun(tags=('ReRunVerifyTag', 'ReRunVerifyOtherTag'), run_upgrade=True)
    _evaluate_results(result, 'ReRunActor', 'ReRunActorOther')


@pytest.mark.skip(reason='Commands have been moved to leapp-repository - The tests will be moved in a separate PR')
def test_leapp_rerun_only_actors_with_not_used_tag():
    result = _perform_rerun(tags=('NoSuchTag',), run_upgrade=True)
    _evaluate_results(result)



import os
import pytest

from leapp.libraries.stdlib import CalledProcessError, run
from leapp.libraries.stdlib.config import is_debug, is_verbose


_STDIN = [[], {}, 0.123, lambda: None]


def test_check_single_line_output():
    cmd = ['echo', 'This a single line test!']
    assert run(cmd, split=True)['stdout'] == [u'This a single line test!']


def test_check_single_line_output_no_split():
    cmd = ['echo', 'This a single line No Split test!']
    assert run(cmd, split=False)['stdout'] == u'This a single line No Split test!\n'


def test_check_multiline_output():
    cmd = ['echo', 'This a multi-\nline test!']
    assert run(cmd, split=True)['stdout'] == [u'This a multi-', u'line test!']


def test_check_multiline_output_no_split():
    cmd = ['echo', 'This a multi-\nline No Split test!']
    assert run(cmd, split=False)['stdout'] == u'This a multi-\nline No Split test!\n'


def test_check_error():
    cmd = ['false']
    with pytest.raises(CalledProcessError):
        run(cmd, checked=True)


def test_check_error_no_checked():
    cmd = ['false']
    assert run(cmd, checked=False)['exit_code'] == 1


def test_stdin_string():
    ret = run(('bash', '-c', 'read MSG; echo "<$MSG>"'), stdin='LOREM IPSUM')
    assert ret['stdout'] == '<LOREM IPSUM>\n'


def test_stdin_fd():
    r, w = os.pipe()
    # The string we write here should not exceed `/proc/sys/fs/pipe-max-size`
    # which represents the size of the kernel buffer backing the pipe
    os.write(w, b'LOREM IPSUM')
    os.close(w)
    ret = run(('bash', '-c', 'read MSG; echo "<$MSG>"'), stdin=r)
    os.close(r)
    assert ret['stdout'] == '<LOREM IPSUM>\n'


@pytest.mark.parametrize('p', _STDIN)
def test_stdin_check(p):
    with pytest.raises(TypeError):
        run(('true',), stdin=p)


def test_is_verbose(monkeypatch):
    matrix = (
        (('1', '1'), True),
        (('0', '1'), True),
        (('1', '0'), True),
        (('0', '0'), False),
        ((0, 1), False),
        ((1, 1), False),
        ((1, 0), False),
        ((0, 0), False),
        (('1', 0), True),
        ((0, '1'), True)
    )
    for (debug, verbose), expected in matrix:
        monkeypatch.setattr(os, 'environ', {'LEAPP_DEBUG': debug, 'LEAPP_VERBOSE': verbose})
        assert is_verbose() == expected


def test_is_debug(monkeypatch):
    matrix = (
        (('1', '1'), True),
        (('0', '1'), False),
        (('1', '0'), True),
        (('0', '0'), False),
        ((0, 1), False),
        ((1, 1), False),
        ((1, 0), False),
        ((0, 0), False),
        (('1', 0), True),
        ((0, '1'), False)
    )
    for (debug, verbose), expected in matrix:
        monkeypatch.setattr(os, 'environ', {'LEAPP_DEBUG': debug, 'LEAPP_VERBOSE': verbose})
        assert is_debug() == expected



from leapp.actors import Actor
from leapp.tags import InplaceUpgradeWorkflowTag, PhaseATag


class PhaseAActor(Actor):
    """
    No documentation has been provided for the phase_a_actor actor.
    """

    name = 'phase_a_actor'
    consumes = ()
    produces = ()
    tags = (InplaceUpgradeWorkflowTag, PhaseATag)

    def process(self):
        print('<<<TEST>>>: {}'.format(self.__class__.__name__))



from leapp.actors import Actor
from leapp.tags import InplaceUpgradeWorkflowTag, PhaseBTag


class PhaseBActor(Actor):
    """
    No documentation has been provided for the phase_b_actor actor.
    """

    name = 'phase_b_actor'
    consumes = ()
    produces = ()
    tags = (InplaceUpgradeWorkflowTag, PhaseBTag)

    def process(self):
        print('<<<TEST>>>: {}'.format(self.__class__.__name__))



from leapp.workflows import Workflow
from leapp.workflows.phases import Phase
from leapp.workflows.flags import Flags
from leapp.workflows.tagfilters import TagFilter
from leapp.workflows.policies import Policies
from leapp.tags import InplaceUpgradeWorkflowTag, PhaseATag, PhaseBTag, FirstBootTag


class IPUWorkflow(Workflow):
    name = 'IPUWorkflow'
    tag = InplaceUpgradeWorkflowTag
    short_name = 'ipu'
    description = '''No description has been provided for the InplaceUpgrade workflow.'''

    # Template for phase definition - The order in which the phase classes are defined
    # within the Workflow class represents the execution order
    #
    # class PhaseName(Phase):
    #    name = 'phase_name'
    #    filter = TagFilter(PhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()

    class PhaseA(Phase):
        name = 'phase_a'
        filter = TagFilter(PhaseATag)
        policies = Policies(Policies.Errors.FailPhase, Policies.Retry.Phase)
        flags = Flags()

    class PhaseB(Phase):
        name = 'phase_b'
        filter = TagFilter(PhaseBTag)
        policies = Policies(Policies.Errors.FailPhase, Policies.Retry.Phase)
        flags = Flags()

    class FirstBoot(Phase):
        name = 'FirstBoot'
        filter = TagFilter(FirstBootTag)
        policies = Policies(Policies.Errors.FailPhase, Policies.Retry.Phase)
        flags = Flags()



import locale
import logging
import os

import leapp.compat


def setup_locale():
    try:
        leapp.compat.setlocale(locale.LC_ALL)
    except locale.Error:
        logging.getLogger('leapp.i18n').error('Failed to set locale, defaulting to C', exc_info=True)
        os.environ['LC_ALL'] = 'C'
        leapp.compat.setlocale(locale.LC_ALL, 'C')


def translation(domain):
    setup_locale()
    return leapp.compat.gettext_setup(
        leapp.compat.gettext.translation(domain, fallback=True))


def install_translation_for_actor(actor):
    if actor.text_domain:
        globals()['__builtins__']['_'], globals()['__builtins__']['P_'] = translation(actor.text_domain)


_, P_ = translation(os.environ.get('LEAPP_DEFAULT_TEXTDOMAIN', 'leapp'))  # noqa; pylint: disable=redefined-builtin

globals()['__builtins__']['_'], globals()['__builtins__']['P_'] = _, P_



import os

from leapp.actors import Actor
from leapp.tags import ThirdPhaseTag, UnitTestWorkflowTag


class BeforeThirdActor(Actor):
    name = 'before_third_actor'
    description = 'No description has been provided for the before_third_actor actor.'
    consumes = ()
    produces = ()
    tags = (ThirdPhaseTag.Before, UnitTestWorkflowTag)

    def process(self):
        from leapp.libraries.common.test_helper import log_execution
        log_execution(self)
        if os.environ.get('BeforeThirdActor-ReportError') == '1':
            self.report_error("Unit test requested error")



import datetime
import json
import hashlib
import logging
import os
import six

from leapp.compat import string_types
from leapp.models import fields, Model, ErrorModel
from leapp.topics import ReportTopic
from leapp.libraries.stdlib.api import produce
from leapp.utils.deprecation import deprecated


class Report(Model):
    """
    Framework model used for reporting
    """

    __non_inheritable__ = True

    topic = ReportTopic

    report = fields.JSON()


def _add_to_dict(data, path, value, leaf_list=False, flat_list=False):
    for p in path[:-1]:
        if p not in data:
            data[p] = {}
        data = data[p]

    if leaf_list:
        list_val = [value]
        if flat_list and isinstance(value, list):
            list_val = value
        data[path[-1]] = data.get(path[-1], []) + list_val
    else:
        if path[-1] in data:
            raise ValueError('Path {} is already taken'.format('.'.join(path)))
        data[path[-1]] = value


class BasePrimitive(object):
    """
    Report primitive (field) base class for dict targets (implies unique path)
    """
    # deepest nested key in resulting report dict
    name = ''

    def __init__(self, value=None):
        if not isinstance(value, string_types):
            raise TypeError('Value of "{}" must be a string'.format(self.__class__.__name__))
        self._value = value

    @property
    def value(self):
        """Return report field value"""
        return self._value

    @property
    def path(self):
        """Return report field destination path (nested dict description)"""
        return (self.name, )

    def to_dict(self):
        return {self.name: self._value}

    def apply(self, report):
        """Add report entry to the report dict based on provided path"""
        _add_to_dict(report, self.path, self.value, leaf_list=False)


class BaseListPrimitive(BasePrimitive):
    """
    Report primitive (field) base class for list targets (we append value to a list)
    """

    def apply(self, report):
        """Add report entry to the report dict based on provided path"""
        _add_to_dict(report, self.path, self.value, leaf_list=True, flat_list=True)


class Title(BasePrimitive):
    """Report title"""
    name = 'title'


class Summary(BasePrimitive):
    """Report summary"""
    name = 'summary'


class Severity(BasePrimitive):
    """Report severity"""
    name = 'severity'

    INFO = 'info'
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'

    def __init__(self, value=None):
        severities = (self.INFO, self.LOW, self.MEDIUM, self.HIGH)
        if value not in severities:
            raise ValueError('Severity should be one of: "{}"'.format(' '.join(severities)))
        self._value = value


class Audience(BasePrimitive):
    """Target audience of the report"""
    name = 'audience'

    def __init__(self, value=None):
        if not isinstance(value, string_types):
            raise TypeError('Value of "Audience" must be a string')
        audiences = ('sysadmin', 'developer')
        if value not in audiences:
            raise ValueError('Audience should be one of: "{}"'.format(' '.join(audiences)))
        self._value = value


class Groups(BaseListPrimitive):
    """Report groups."""
    name = 'groups'

    INHIBITOR = 'inhibitor'
    FAILURE = 'failure'
    ACCESSIBILITY = 'accessibility'
    AUTHENTICATION = 'authentication'
    BOOT = 'boot'
    COMMUNICATION = 'communication'
    DESKTOP = 'desktop environment'
    DRIVERS = 'drivers'
    EMAIL = 'email'
    ENCRYPTION = 'encryption'
    FILESYSTEM = 'filesystem'
    FIREWALL = 'firewall'
    HIGH_AVAILABILITY = 'high availability'
    KERNEL = 'kernel'
    MONITORING = 'monitoring'
    NETWORK = 'network'
    OS_FACTS = 'OS facts'
    POST = 'post'
    PUBLIC_CLOUD = 'public cloud'
    PYTHON = 'python'
    REPOSITORY = 'repository'
    RHUI = 'rhui'
    SANITY = 'sanity'
    SECURITY = 'security'
    SELINUX = 'selinux'
    SERVICES = 'services'
    TIME_MANAGEMENT = 'time management'
    TOOLS = 'tools'
    UPGRADE_PROCESS = 'upgrade process'

    def __init__(self, value=None):
        if not isinstance(value, list):
            raise TypeError('Value of "Groups" must be a list')
        self._value = value


@deprecated(
    since='2022-08-23',
    message=(
        'The primitive is deprecated as Tags and Flags have been joined into the Groups primitive.'
        'Please use Groups for report message typing instead.'
    )
)
class Tags(Groups):
    def __init__(self, *args, **kwargs):
        super(Tags, self).__init__(*args, **kwargs)


@deprecated(
    since='2022-08-23',
    message=(
        'The primitive is deprecated as Tags and Flags have been joined into the Groups primitive.'
        'Please use Groups for report message typing instead.'
    )
)
class Flags(Groups):
    def __init__(self, *args, **kwargs):
        super(Flags, self).__init__(*args, **kwargs)


# To allow backwards-compatibility with previous report-schema
# Groups that match _DEPRECATION_FLAGS will be shown as flags, the rest as tags
_DEPRECATION_FLAGS = [Groups.INHIBITOR, Groups.FAILURE]


class Key(BasePrimitive):
    """Stable identifier for report entries."""
    name = 'key'

    def __init__(self, uuid):
        # uuid is allowed to be string or unicode only, checking in py2/py3 friendly mode
        if not isinstance(uuid, six.string_types):
            raise ValueError('Key value should be a string.')

        self._value = uuid


class ExternalLink(BaseListPrimitive):
    """External link report detail field"""
    name = 'external'

    def __init__(self, url=None, title=None):
        if not all(isinstance(v, string_types) for v in (url, title)):
            raise TypeError('Values "url" and "title" of "ExternalLink" must be a string')
        self._value = {'url': url, 'title': title}

    @property
    def path(self):
        return ('detail', 'external')


class RelatedResource(BaseListPrimitive):
    """Report detail field for related resources (e.g. affected packages/files)"""
    name = 'related_resources'

    def __init__(self, scheme=None, identifier=None):
        if not all(isinstance(v, string_types) for v in (scheme, identifier)):
            raise TypeError('Values "scheme" and "identifier" of "RelatedResource" must be a string')
        self._value = {'scheme': scheme, 'title': identifier}

    @property
    def path(self):
        return ('detail', 'related_resources')


class BaseRemediation(BaseListPrimitive):
    """Remediation base class"""
    name = 'remediations'

    @property
    def path(self):
        return ('detail', 'remediations')


class RemediationCommand(BaseRemediation):
    def __init__(self, value=None):
        if not isinstance(value, list):
            raise TypeError('Value of "RemediationCommand" must be a list')
        self._value = {'type': 'command', 'context': value}

    def __repr__(self):
        return "[{}] {}".format(self._value['type'], ' '.join(self._value['context']))


class RemediationHint(BaseRemediation):
    def __init__(self, value=None):
        self._value = {'type': 'hint', 'context': value}

    def __repr__(self):
        return "[{}] {}".format(self._value['type'], self._value['context'])


class RemediationPlaybook(BaseRemediation):
    def __init__(self, value=None):
        self._value = {'type': 'playbook', 'context': value}

    def __repr__(self):
        return "[{}] {}".format(self._value['type'], self._value['context'])


class Remediation(object):
    def __init__(self, playbook=None, commands=None, hint=None):
        self._remediations = []
        if hint:
            self._remediations.append(RemediationHint(value=hint))
        if commands:
            for command in commands:
                self._remediations.append(RemediationCommand(value=command))
        if playbook:
            self._remediations.append(RemediationPlaybook(value=playbook))

    def apply(self, report):
        for remediation in self._remediations:
            remediation.apply(report)

    def __repr__(self):
        return "\n".join([repr(r) for r in self._remediations])

    def to_dict(self):
        return {'remediations': [r.to_dict() for r in self._remediations]}

    @classmethod
    def from_dict(cls, data):
        values = data.get('remediations', [])
        playbook = next((v['context'] for v in values if v['type'] == 'playbook'), None)
        hint = next((v['context'] for v in values if v['type'] == 'hint'), None)
        commands = [v['context'] for v in values if v['type'] == 'command']
        if values:
            return Remediation(playbook, commands, hint)


def _sanitize_entries(entries):
    if not any(isinstance(e, Title) for e in entries):
        raise ValueError('Title not provided in the report.')
    if not any(isinstance(e, Summary) for e in entries):
        raise ValueError('Summary not provided in the report.')
    if not any(isinstance(e, Severity) for e in entries):
        entries.append(Severity('info'))
    if not any(isinstance(e, Audience) for e in entries):
        entries.append(Audience('sysadmin'))
    _check_stable_key(entries, raise_if_undefined=os.getenv('LEAPP_DEVEL_FIXED_REPORT_KEY', '0') != '0')


def _check_stable_key(entries, raise_if_undefined=False):
    """
    Inserts a Key into the entries of the report if it doesn't exist.

    The resulting id is calculated from the mandatory report
    entries (severity, audience, title).
    Please note that the original entries list is modified.
    """
    def _find_entry(cls):
        entry = next((e for e in entries if isinstance(e, cls)), None)
        if entry:
            return entry._value

    logger = logging.getLogger('leapp.reporting')
    key = _find_entry(Key)
    if key is not None:
        # Key is already there, check that it's not an empty string
        if not key.strip():
            logger.error('An empty string is specified as Key')
            raise ValueError('Key can not be an empty string.')
        return

    if raise_if_undefined:
        raise ValueError('Stable Key report entry with specified unique value not provided')

    # No Key found - let's generate one!
    key_str = u'{severity}:{audience}:{title}'.format(severity=_find_entry(Severity),
                                                      audience=_find_entry(Audience),
                                                      title=_find_entry(Title))
    key_value = hashlib.sha1(key_str.encode('utf-8')).hexdigest()
    entries.append(Key(key_value))
    logger.warning('Stable Key report entry not provided, dynamically generating one - {}'.format(key_value))


def _create_report_object(entries):
    report = {}

    _sanitize_entries(entries)
    for entry in entries:
        entry.apply(report)
    if Groups.INHIBITOR in report.get('groups', []):
        # Before removing Flags, the original inhibitor detection worked
        # by checking the `flags` field; keep the `flags` field until we drop Flags completely
        # Currently we know that 'flags' does not exist otherwise so it's
        # safe to just set it
        report['flags'] = [Groups.INHIBITOR]

    return Report(report=report)


def create_report(entries):
    """
    Produce final report message
    """
    produce(_create_report_object(entries))


_DEPRECATION_SEVERITY_THRESHOLD = datetime.timedelta(days=180)


def create_report_from_deprecation(deprecation):
    """
    Convert deprecation json to report json
    """
    since = datetime.datetime.strptime(deprecation['since'], '%Y-%m-%d').date()

    severity = Severity.MEDIUM
    if since < (datetime.date.today() - _DEPRECATION_SEVERITY_THRESHOLD):
        severity = Severity.HIGH

    report = _create_report_object([
        Title('{message} at {filename}:{lineno}'.format(
            message=deprecation['message'].split('\n')[0],
            filename=deprecation['filename'],
            lineno=deprecation['lineno'])),
        Severity(severity),
        Audience('developer'),
        Summary('{reason}\nSince: {since}\nLocation: {filename}:{lineno}\nNear: {line}'.format(
            reason=deprecation['reason'],
            since=deprecation['since'],
            filename=deprecation['filename'],
            lineno=deprecation['lineno'],
            line=deprecation['line']
        )),
    ])
    return json.loads(report.dump()['report'])


def create_report_from_error(error_dict):
    """
    Convert error json to report json
    """
    error = ErrorModel.create(error_dict)
    entries = [Title(error.message), Summary(error.details or ""), Severity('high'), Audience('sysadmin')]
    report = _create_report_object(entries)
    return json.loads(report.dump()['report'])
